#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The messages API."""

from http import HTTPStatus

import webob

from cinder.api import common
from cinder.api import microversions as mv
from cinder.api.openstack import wsgi
from cinder.api.v3.views import messages as messages_view
from cinder.message import api as message_api
from cinder.message import defined_messages
from cinder.message import message_field
from cinder.policies import messages as policy


class MessagesController(wsgi.Controller):
    """The User Messages API controller for the OpenStack API."""

    _view_builder_class = messages_view.ViewBuilder

    def __init__(self, ext_mgr):
        self.message_api = message_api.API()
        self.ext_mgr = ext_mgr
        super(MessagesController, self).__init__()

    def _build_user_message(self, message):
        # NOTE(tommylikehu): if the `action_id` is empty, we use 'event_id'
        # to translate the user message.
        if message is None:
            return
        if message['action_id'] is None and message['event_id'] is not None:
            message['user_message'] = defined_messages.get_message_text(
                message['event_id'])
        else:
            message['user_message'] = "%s:%s" % (
                message_field.translate_action(message['action_id']),
                message_field.translate_detail(message['detail_id']))

    @wsgi.Controller.api_version(mv.MESSAGES)
    def show(self, req, id):
        """Return the given message."""
        context = req.environ['cinder.context']

        # Not found exception will be handled at the wsgi level
        message = self.message_api.get(context, id)

        context.authorize(policy.GET_POLICY, target_obj=message)

        self._build_user_message(message)
        return self._view_builder.detail(req, message)

    @wsgi.Controller.api_version(mv.MESSAGES)
    def delete(self, req, id):
        """Delete a message."""
        context = req.environ['cinder.context']

        # Not found exception will be handled at the wsgi level
        message = self.message_api.get(context, id)
        context.authorize(policy.DELETE_POLICY, target_obj=message)
        self.message_api.delete(context, message)

        return webob.Response(status_int=HTTPStatus.NO_CONTENT)

    @wsgi.Controller.api_version(mv.MESSAGES)
    def index(self, req):
        """Returns a list of messages, transformed through view builder."""
        context = req.environ['cinder.context']
        api_version = req.api_version_request
        context.authorize(policy.GET_ALL_POLICY)
        filters = None
        marker = None
        limit = None
        offset = None
        sort_keys = None
        sort_dirs = None

        if api_version.matches(mv.MESSAGES_PAGINATION):
            filters = req.params.copy()
            marker, limit, offset = common.get_pagination_params(filters)
            sort_keys, sort_dirs = common.get_sort_params(filters)

        if api_version.matches(mv.RESOURCE_FILTER):
            support_like = (True if api_version.matches(
                mv.LIKE_FILTER) else False)
            common.reject_invalid_filters(context, filters, 'message',
                                          support_like)

        messages = self.message_api.get_all(context, filters=filters,
                                            marker=marker, limit=limit,
                                            offset=offset,
                                            sort_keys=sort_keys,
                                            sort_dirs=sort_dirs)

        for message in messages:
            self._build_user_message(message)
        messages = self._view_builder.index(req, messages)
        return messages


def create_resource(ext_mgr):
    return wsgi.Resource(MessagesController(ext_mgr))



# Copyright 2015 IBM Corp.
# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""
Volume FC driver for IBM Storwize family and SVC storage systems.

Notes:
1. If you specify both a password and a key file, this driver will use the
key file only.
2. When using a key file for authentication, it is up to the user or
system administrator to store the private key in a safe manner.
3. The defaults for creating volumes are "-rsize 2% -autoexpand
-grainsize 256 -warning 0".  These can be changed in the configuration
file or by using volume types(recommended only for advanced users).

Limitations:
1. The driver expects CLI output in English, error messages may be in a
localized format.
2. Clones and creating volumes from snapshots, where the source and target
are of different sizes, is not supported.

"""
import collections

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from cinder import coordination
from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder.volume import configuration
from cinder.volume.drivers.ibm.storwize_svc import (
    storwize_svc_common as storwize_common)
from cinder.zonemanager import utils as fczm_utils

LOG = logging.getLogger(__name__)

storwize_svc_fc_opts = [
    cfg.BoolOpt('storwize_svc_multipath_enabled',
                default=False,
                help='Connect with multipath (FC only; iSCSI multipath is '
                     'controlled by Nova)'),
]

CONF = cfg.CONF
CONF.register_opts(storwize_svc_fc_opts, group=configuration.SHARED_CONF_GROUP)


@interface.volumedriver
class StorwizeSVCFCDriver(storwize_common.StorwizeSVCCommonDriver):
    """IBM Storwize V7000 and SVC FC volume driver.

    Version history:

    .. code-block:: none

        1.0 - Initial driver
        1.1 - FC support, create_cloned_volume, volume type support,
              get_volume_stats, minor bug fixes
        1.2.0 - Added retype
        1.2.1 - Code refactor, improved exception handling
        1.2.2 - Fix bug #1274123 (races in host-related functions)
        1.2.3 - Fix Fibre Channel connectivity: bug #1279758 (add delim
                to lsfabric, clear unused data from connections, ensure
                matching WWPNs by comparing lower case
        1.2.4 - Fix bug #1278035 (async migration/retype)
        1.2.5 - Added support for manage_existing (unmanage is inherited)
        1.2.6 - Added QoS support in terms of I/O throttling rate
        1.3.1 - Added support for volume replication
        1.3.2 - Added support for consistency group
        1.3.3 - Update driver to use ABC metaclasses
        2.0 - Code refactor, split init file and placed shared methods
              for FC and iSCSI within the StorwizeSVCCommonDriver class
        2.0.1 - Added support for multiple pools with model update
        2.1 - Added replication V2 support to the global/metro mirror
              mode
        2.1.1 - Update replication to version 2.1
        2.2 - Add CG capability to generic volume groups
        2.2.1 - Add vdisk mirror/stretch cluster support
        2.2.2 - Add npiv support
        2.2.3 - Add replication group support
        2.2.4 - Add backup snapshots support
        2.2.5 - Add hyperswap support
        2.2.6 - Add support for host attachment using portsets
    """

    VERSION = "2.2.6"

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "IBM_STORAGE_CI"

    def __init__(self, *args, **kwargs):
        super(StorwizeSVCFCDriver, self).__init__(*args, **kwargs)
        self.protocol = 'FC'
        self.configuration.append_config_values(
            storwize_svc_fc_opts)

    @staticmethod
    def get_driver_options():
        return storwize_common.storwize_svc_opts + storwize_svc_fc_opts

    def validate_connector(self, connector):
        """Check connector for at least one enabled FC protocol."""
        if 'wwpns' not in connector:
            LOG.error('The connector does not contain the required '
                      'information.')
            raise exception.InvalidConnectorException(
                missing='wwpns')

    def initialize_connection_snapshot(self, snapshot, connector):
        """Perform attach snapshot for backup snapshots."""
        # If the snapshot's source volume is a replication volume and the
        # replication volume has failed over to aux_backend,
        # attach the snapshot will be failed.
        self._check_snapshot_replica_volume_status(snapshot)

        vol_attrs = ['id', 'name', 'volume_type_id', 'display_name']
        Volume = collections.namedtuple('Volume', vol_attrs)
        volume = Volume(id=snapshot.id,
                        name=snapshot.name,
                        volume_type_id=snapshot.volume_type_id,
                        display_name='backup-snapshot')

        return self.initialize_connection(volume, connector)

    def initialize_connection(self, volume, connector):
        """Perform necessary work to make a FC connection."""
        @coordination.synchronized('storwize-host-{system_id}-{host}')
        def _do_initialize_connection_locked(system_id, host):
            conn_info = self._do_initialize_connection(volume, connector)
            fczm_utils.add_fc_zone(conn_info)
            return conn_info
        return _do_initialize_connection_locked(self._state['system_id'],
                                                connector['host'])

    def _do_initialize_connection(self, volume, connector):
        """Perform necessary work to make a FC connection.

        To be able to create an FC connection from a given host to a
        volume, we must:
        1. Translate the given WWNN to a host name
        2. Create new host on the storage system if it does not yet exist
        3. Map the volume to the host if it is not already done
        4. Return the connection information for relevant nodes (in the
           proper I/O group)

        """
        LOG.debug('enter: initialize_connection: volume %(vol)s with connector'
                  ' %(conn)s', {'vol': volume.id, 'conn': connector})
        if volume.display_name == 'backup-snapshot':
            LOG.debug('It is a virtual volume %(vol)s for attach snapshot.',
                      {'vol': volume.id})
            volume_name = volume.name
            backend_helper = self._helpers
            node_state = self._state
        else:
            volume_name, backend_helper, node_state = self._get_vol_sys_info(
                volume)

        host_site = None
        backend_helper.initialize_host_info()

        is_hyper_volume = self.is_volume_hyperswap(volume)
        if is_hyper_volume:
            host_site = self._get_volume_host_site_from_conf(volume,
                                                             connector)
        # The host_site is necessary for hyperswap volume.
        if is_hyper_volume and host_site is None:
            msg = (_('There is no correct storwize_preferred_host_site '
                     'configured for a hyperswap volume %s.') % volume.name)
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        # Try creating the host, if host creation is successfull continue
        # with intialization flow, else search for host object defined for
        # this connector info.
        host_name = None
        try:
            opts = self._get_vdisk_params(volume.volume_type_id)
            host_name = (
                backend_helper.create_host(connector, site=host_site,
                                           portset=opts['storwize_portset']))
        except exception.VolumeBackendAPIException as excp:
            if "CMMVC6035E" in excp.msg:
                msg = (_('Host already exists for connector '
                         '%(conn)s'), {'conn': connector})
                LOG.info(msg)
                host_name = backend_helper.get_host_from_connector(connector)
            else:
                msg = (_('Error creating host %(ex)s'), {'ex': excp.msg})
                LOG.error(msg)
                raise exception.VolumeDriverException(message=msg)

        if is_hyper_volume:
            self._update_host_site_for_hyperswap_volume(host_name, host_site)

        volume_attributes = backend_helper.get_vdisk_attributes(volume_name)
        if volume_attributes is None:
            msg = (_('initialize_connection: Failed to get attributes'
                     ' for volume %s.') % volume_name)
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        multihostmap = self.configuration.storwize_svc_multihostmap_enabled
        lun_id = backend_helper.map_vol_to_host(volume_name, host_name,
                                                multihostmap)
        try:
            preferred_node = volume_attributes['preferred_node_id']
            IO_group = volume_attributes['IO_group_id']
        except KeyError as e:
            LOG.error('Did not find expected column name in '
                      'lsvdisk: %s.', e)
            raise exception.VolumeBackendAPIException(
                data=_('initialize_connection: Missing volume attribute for '
                       'volume %s.') % volume_name)

        try:
            # Get preferred node and other nodes in I/O group
            preferred_node_entry = None
            io_group_nodes = []
            for node in node_state['storage_nodes'].values():
                if node['id'] == preferred_node:
                    preferred_node_entry = node
                if node['IO_group'] == IO_group:
                    io_group_nodes.append(node)

            if not len(io_group_nodes):
                msg = (_('initialize_connection: No node found in '
                         'I/O group %(gid)s for volume %(vol)s.') %
                       {'gid': IO_group, 'vol': volume_name})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            if not preferred_node_entry:
                # Get 1st node in I/O group
                preferred_node_entry = io_group_nodes[0]
                LOG.warning('initialize_connection: Did not find a '
                            'preferred node for volume %s.', volume_name)

            properties = {}
            properties['target_discovered'] = False
            properties['target_lun'] = lun_id
            properties['volume_id'] = volume.id

            conn_wwpns = backend_helper.get_conn_fc_wwpns(host_name)

            # If conn_wwpns is empty, then that means that there were
            # no target ports with visibility to any of the initiators
            # so we return all target ports.
            if len(conn_wwpns) == 0:
                for node in node_state['storage_nodes'].values():
                    # The Storwize/svc release 7.7.0.0 introduced NPIV feature,
                    # Different commands be used to get the wwpns for host I/O
                    if node_state['code_level'] < (7, 7, 0, 0):
                        conn_wwpns.extend(node['WWPN'])
                    else:
                        npiv_wwpns = backend_helper.get_npiv_wwpns(
                            node_state['code_level'],
                            node_id=node['id'],
                            host_io="yes",
                            portset=opts['storwize_portset'])
                        conn_wwpns.extend(npiv_wwpns)

            properties['target_wwn'] = conn_wwpns

            i_t_map = self._make_initiator_target_map(connector['wwpns'],
                                                      conn_wwpns)
            properties['initiator_target_map'] = i_t_map

            # specific for z/VM, refer to cinder bug 1323993
            if "zvm_fcp" in connector:
                properties['zvm_fcp'] = connector['zvm_fcp']
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error('initialize_connection: Failed to export volume '
                          '%(vol)s due to %(ex)s.', {'vol': volume.name,
                                                     'ex': ex})
                self._do_terminate_connection(volume, connector)
                LOG.error('initialize_connection: Failed '
                          'to collect return '
                          'properties for volume %(vol)s and connector '
                          '%(conn)s.\n', {'vol': volume,
                                          'conn': connector})

        LOG.debug('leave: initialize_connection:\n volume: %(vol)s\n '
                  'connector %(conn)s\n properties: %(prop)s',
                  {'vol': volume.id, 'conn': connector,
                   'prop': properties})

        return {'driver_volume_type': 'fibre_channel', 'data': properties, }

    def _make_initiator_target_map(self, initiator_wwpns, target_wwpns):
        """Build a simplistic all-to-all mapping."""
        i_t_map = {}
        for i_wwpn in initiator_wwpns:
            i_t_map[str(i_wwpn)] = []
            for t_wwpn in target_wwpns:
                i_t_map[i_wwpn].append(t_wwpn)

        return i_t_map

    def terminate_connection_snapshot(self, snapshot, connector, **kwargs):
        """Perform detach snapshot for backup snapshots."""
        vol_attrs = ['id', 'name', 'display_name']
        Volume = collections.namedtuple('Volume', vol_attrs)
        volume = Volume(id=snapshot.id,
                        name=snapshot.name,
                        display_name='backup-snapshot')

        return self.terminate_connection(volume, connector, **kwargs)

    def terminate_connection(self, volume, connector, **kwargs):
        """Cleanup after an FC connection has been terminated."""
        # If a fake connector is generated by nova when the host
        # is down, then the connector will not have a host property,
        # In this case construct the lock without the host property
        # so that all the fake connectors to an SVC are serialized
        host = connector['host'] if 'host' in connector else ""
        attachment_count = 0
        if hasattr(volume, 'multiattach') and volume.multiattach:
            try:
                attachment_list = volume.volume_attachment
                for attachment in attachment_list:
                    if (attachment.attach_status == "attached" and
                       attachment.attached_host == host):
                        attachment_count += 1
            except AttributeError:
                pass
            if attachment_count > 1:
                LOG.debug("Volume %(volume)s is attached to multiple "
                          "instances on host %(host_name)s, "
                          "skip terminate volume connection",
                          {'volume': volume.name,
                           'host_name': volume.host.split('@')[0]})
                return

        @coordination.synchronized('storwize-host-{system_id}-{host}')
        def _do_terminate_connection_locked(system_id, host):
            conn_info = self._do_terminate_connection(volume, connector,
                                                      **kwargs)
            fczm_utils.remove_fc_zone(conn_info)
            return conn_info
        return _do_terminate_connection_locked(self._state['system_id'], host)

    def _do_terminate_connection(self, volume, connector, **kwargs):
        """Cleanup after an FC connection has been terminated.

        When we clean up a terminated connection between a given connector
        and volume, we:
        1. Translate the given connector to a host name
        2. Remove the volume-to-host mapping if it exists
        3. Delete the host if it has no more mappings (hosts are created
           automatically by this driver when mappings are created)
        """
        LOG.debug('enter: terminate_connection: volume %(vol)s with connector'
                  ' %(conn)s', {'vol': volume.id, 'conn': connector})
        (info, host_name, vol_name, backend_helper,
         node_state) = self._get_map_info_from_connector(volume, connector)

        if not backend_helper:
            return info

        # Unmap volumes, if hostname is None, need to get value from vdiskmap
        host_name = backend_helper.unmap_vol_from_host(vol_name, host_name)

        # Host_name could be none
        if host_name:
            resp = backend_helper.check_host_mapped_vols(host_name)
            if not len(resp):
                LOG.info("Need to remove FC Zone, building initiator "
                         "target map.")
                # Build info data structure for zone removing
                if 'wwpns' in connector and host_name:
                    target_wwpns = []
                    # Returning all target_wwpns in storage_nodes, since
                    # we cannot determine which wwpns are logged in during
                    # a VM deletion.
                    for node in node_state['storage_nodes'].values():
                        target_wwpns.extend(node['WWPN'])
                    init_targ_map = (self._make_initiator_target_map
                                     (connector['wwpns'],
                                      target_wwpns))
                    info['data'] = {'initiator_target_map': init_targ_map}
                # No volume mapped to the host, delete host from array
                backend_helper.delete_host(host_name)

        LOG.debug('leave: terminate_connection: volume %(vol)s with '
                  'connector %(conn)s, info %(info)s', {'vol': volume.id,
                                                        'conn': connector,
                                                        'info': info})
        return info

    def _get_volume_connection_info(self, ctxt, volume, host_info,
                                    iogrp_list):
        connector = {'wwpns': []}
        connection_info = {"driver_volume_type": "fibre_channel"}
        data = {}

        for wwpn in host_info.select('WWPN'):
            connector['wwpns'].append(wwpn)

        vol_name, backend_helper, node_state = self._get_vol_sys_info(volume)

        data['target_discovered'] = False
        data['volume_id'] = volume.id

        conn_wwpns = []
        for node in node_state['storage_nodes'].values():
            if node['IO_group'] not in iogrp_list:
                continue

            # The Storwize/svc release 7.7.0.0 introduced NPIV feature,
            # Different commands be used to get the wwpns for host I/O
            if node_state['code_level'] < (7, 7, 0, 0):
                conn_wwpns.extend(node['WWPN'])
            else:
                npivwwpns = (
                    backend_helper.get_npiv_wwpns(node_state['code_level'],
                                                  node_id=node['id'],
                                                  host_io="yes"))
                conn_wwpns.extend(npivwwpns)

        i_t_map = self._make_initiator_target_map(connector['wwpns'],
                                                  conn_wwpns)
        data["initiator_target_map"] = i_t_map
        data["target_wwn"] = conn_wwpns

        connection_info['data'] = data
        connection_info['connector'] = connector

        return connection_info

    def _retype_hyperswap_volume(self, ctxt, volume, host, old_opts,
                                 new_opts, old_pool, new_pool, vdisk_changes,
                                 need_copy, new_type):
        if (old_opts['volume_topology'] != 'hyperswap' and
                new_opts['volume_topology'] == 'hyperswap'):
            LOG.debug('retype: Convert a normal volume %s to hyperswap '
                      'volume.', volume.name)
            conn_info = {}
            if volume.previous_status == 'in-use':
                vdisk_info = self._helpers.ssh.lsvdiskhostmap(volume.name)
                peer_pool = new_opts['peer_pool']
                iogrp_list = self._helpers.get_hyperswap_pool_io_grp(
                    self._state, new_pool, peer_pool)
                for mapping_info in vdisk_info:
                    host = mapping_info['host_name']
                    try:
                        host_info = self._helpers.ssh.lshost(host)
                        conn_info[host] = self._get_volume_connection_info(
                            ctxt, volume, host_info, iogrp_list)
                        host_site = self._get_volume_host_site_from_conf(
                            volume, conn_info[host].get('connector'))
                        self._update_host_site_for_hyperswap_volume(
                            host, host_site)
                        self._helpers.ssh.addhostiogrp(host,
                                                       iogrp_list)
                    except Exception as ex:
                        msg = _('Error updating host %(host)s due to %(ex)s',
                                {'host': host, 'ex': ex})
                        raise exception.VolumeBackendAPIException(data=msg)
            self._helpers.convert_volume_to_hyperswap(volume.name,
                                                      new_opts,
                                                      self._state)
            if volume.previous_status == 'in-use':
                for host, info in conn_info.items():
                    try:
                        fczm_utils.add_fc_zone(info)
                    except Exception as ex:
                        self._helpers.convert_hyperswap_volume_to_normal(
                            volume.name, new_opts['peer_pool'])
                        msg = _('Zoning failed for volume %(vol)s and host '
                                '%(host)s due to %(ex)s.',
                                {'vol': volume.name, 'host': host, 'ex': ex})
                        raise exception.VolumeBackendAPIException(data=msg)
        elif (old_opts['volume_topology'] == 'hyperswap' and
                new_opts['volume_topology'] != 'hyperswap'):
            LOG.debug('retype: Convert a hyperswap volume %s to normal '
                      'volume.', volume.name)
            if new_pool == old_pool:
                self._helpers.convert_hyperswap_volume_to_normal(
                    volume.name,
                    old_opts['peer_pool'])
            elif new_pool == old_opts['peer_pool']:
                self._helpers.convert_hyperswap_volume_to_normal(
                    volume.name,
                    old_pool)
            if volume.previous_status == 'in-use':
                vdisk_info = self._helpers.ssh.lsvdiskhostmap(volume.name)
                for mapping_info in vdisk_info:
                    res = self._helpers.check_host_mapped_vols(
                        mapping_info['host_name'])
                    if len(res) == 1:
                        self._helpers.update_host(mapping_info['host_name'],
                                                  None)
        else:
            rel_info = self._helpers.get_relationship_info(volume.name)
            aux_vdisk = rel_info['aux_vdisk_name']
            if need_copy:
                self.add_vdisk_copy(aux_vdisk, old_opts['peer_pool'], new_type,
                                    auto_delete=True)
            elif vdisk_changes:
                self._helpers.change_vdisk_options(aux_vdisk,
                                                   vdisk_changes,
                                                   new_opts, self._state)
        if need_copy:
            self.add_vdisk_copy(volume.name, old_pool, new_type,
                                auto_delete=True)
        elif vdisk_changes:
            self._helpers.change_vdisk_options(volume.name,
                                               vdisk_changes,
                                               new_opts, self._state)



# Copyright 2015 Clinton Knight
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import ddt

from cinder.api.views import versions
from cinder.tests.unit import test


class FakeRequest(object):
    def __init__(self, application_url):
        self.application_url = application_url


URL_BASE = 'http://localhost/volume/'
URL_BASE_NO_SLASH = 'http://localhost/volume'
FAKE_HREF = URL_BASE + 'v1/'

FAKE_VERSIONS = {
    "v1.0": {
        "id": "v1.0",
        "status": "CURRENT",
        "version": "1.1",
        "min_version": "1.0",
        "updated": "2015-07-30T11:33:21Z",
        "links": [
            {
                "rel": "describedby",
                "type": "text/html",
                "href": 'http://docs.openstack.org/',
            },
        ],
        "media-types": [
            {
                "base": "application/json",
                "type": "application/vnd.openstack.share+json;version=1",
            }
        ],
    },
}

FAKE_LINKS = [
    {
        "rel": "describedby",
        "type": "text/html",
        "href": 'http://docs.openstack.org/',
    },
    {
        'rel': 'self',
        'href': FAKE_HREF
    },
]


@ddt.ddt
class ViewBuilderTestCase(test.TestCase):

    def _get_builder(self):
        request = FakeRequest(URL_BASE)
        return versions.get_view_builder(request)

    def _get_builder_no_slash(self):
        request = FakeRequest(URL_BASE_NO_SLASH)
        return versions.get_view_builder(request)

    def test_build_versions(self):

        self.mock_object(versions.ViewBuilder,
                         '_build_links',
                         return_value=FAKE_LINKS)

        fake_versions = copy.deepcopy(FAKE_VERSIONS)
        result = self._get_builder().build_versions(fake_versions)
        result_no_slash = self._get_builder_no_slash().build_versions(
            fake_versions)

        expected = {'versions': list(fake_versions.values())}
        expected['versions'][0]['links'] = FAKE_LINKS

        self.assertEqual(expected, result)
        self.assertEqual(expected, result_no_slash)

    def test_build_version(self):

        self.mock_object(versions.ViewBuilder,
                         '_build_links',
                         return_value=FAKE_LINKS)

        result = self._get_builder()._build_version(FAKE_VERSIONS['v1.0'])
        result_no_slash = self._get_builder_no_slash()._build_version(
            FAKE_VERSIONS['v1.0'])

        expected = copy.deepcopy(FAKE_VERSIONS['v1.0'])
        expected['links'] = FAKE_LINKS

        self.assertEqual(expected, result)
        self.assertEqual(expected, result_no_slash)

    def test_build_links(self):

        self.mock_object(versions.ViewBuilder,
                         '_generate_href',
                         return_value=FAKE_HREF)

        result = self._get_builder()._build_links(FAKE_VERSIONS['v1.0'])
        result_no_slash = self._get_builder_no_slash()._build_links(
            FAKE_VERSIONS['v1.0'])

        self.assertEqual(FAKE_LINKS, result)
        self.assertEqual(FAKE_LINKS, result_no_slash)

    def test_generate_href_defaults(self):

        result = self._get_builder()._generate_href()
        result_no_slash = self._get_builder_no_slash()._generate_href()

        self.assertEqual(URL_BASE + 'v3/', result)
        self.assertEqual(URL_BASE + 'v3/', result_no_slash)

    @ddt.data(
        ('v2', None, URL_BASE + 'v2/'),
        ('/v2/', None, URL_BASE + 'v2/'),
        ('/v2/', 'fake_path', URL_BASE + 'v2/fake_path'),
        ('/v2/', '/fake_path/', URL_BASE + 'v2/fake_path/'),
    )
    @ddt.unpack
    def test_generate_href_no_path(self, version, path, expected):

        result = self._get_builder()._generate_href(version=version,
                                                    path=path)
        result_no_slash = self._get_builder_no_slash()._generate_href(
            version=version, path=path)

        self.assertEqual(expected, result)
        self.assertEqual(expected, result_no_slash)

    @ddt.data(
        ('http://1.1.1.1/', 'http://1.1.1.1/'),
        ('http://localhost/', 'http://localhost/'),
        ('http://localhost/volume/', 'http://localhost/volume/'),
        ('http://1.1.1.1/v1/', 'http://1.1.1.1/'),
        ('http://1.1.1.1/volume/v1/', 'http://1.1.1.1/volume/'),
        ('http://1.1.1.1/v1', 'http://1.1.1.1/'),
        ('http://1.1.1.1/v11', 'http://1.1.1.1/'),
    )
    @ddt.unpack
    def test_get_base_url_without_version(self, base_url, base_url_no_version):

        request = FakeRequest(base_url)
        builder = versions.get_view_builder(request)

        result = builder._get_base_url_without_version()

        self.assertEqual(base_url_no_version, result)



# Copyright 2013 IBM Corp.
# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
IBM Storage driver is a unified Volume driver for IBM XIV, Spectrum Accelerate,
FlashSystem A9000, FlashSystem A9000R and DS8000 storage systems.
"""

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils

from cinder import exception
from cinder import interface
from cinder.volume import configuration
from cinder.volume import driver
from cinder.volume.drivers.san import san
from cinder.zonemanager import utils as fczm_utils

driver_opts = [
    cfg.StrOpt(
        'proxy',
        default='cinder.volume.drivers.ibm.ibm_storage.proxy.IBMStorageProxy',
        help='Proxy driver that connects to the IBM Storage Array'),
    cfg.StrOpt(
        'connection_type',
        default='iscsi',
        choices=['fibre_channel', 'iscsi'],
        help='Connection type to the IBM Storage Array'),
    cfg.StrOpt(
        'chap',
        default='disabled',
        choices=['disabled', 'enabled'],
        help='CHAP authentication mode, effective only for iscsi'
        ' (disabled|enabled)'),
    cfg.StrOpt(
        'management_ips',
        default='',
        help='List of Management IP addresses (separated by commas)'),
]

CONF = cfg.CONF
CONF.register_opts(driver_opts, group=configuration.SHARED_CONF_GROUP)

LOG = logging.getLogger(__name__)


@interface.volumedriver
class IBMStorageDriver(san.SanDriver,
                       driver.ManageableVD,
                       driver.MigrateVD,
                       driver.CloneableImageVD):
    """IBM Storage driver

    IBM Storage driver is a unified Volume driver for IBM XIV, Spectrum
    Accelerate, FlashSystem A9000, FlashSystem A9000R and DS8000 storage
    systems.

    Version history:

    .. code-block:: none

        2.0 - First open source driver version
        2.1.0 - Support Consistency groups through Generic volume groups
              - Support XIV/A9000 Volume independent QoS
              - Support Consistency groups replication
        2.3.0 - Support Report backend state
    """

    VERSION = "2.3.0"

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "IBM_STORAGE_CI"

    def __init__(self, *args, **kwargs):
        """Initialize the driver."""

        super(IBMStorageDriver, self).__init__(*args, **kwargs)

        self.configuration.append_config_values(driver_opts)

        proxy = importutils.import_class(self.configuration.proxy)

        active_backend_id = kwargs.get('active_backend_id', None)

        # Driver additional flags should be specified in the cinder.conf
        # preferably in each backend configuration.

        self.proxy = proxy(
            {
                "user": self.configuration.san_login,
                "password": self.configuration.san_password,
                "address": self.configuration.san_ip,
                "vol_pool": self.configuration.san_clustername,
                "connection_type": self.configuration.connection_type,
                "chap": self.configuration.chap,
                "management_ips": self.configuration.management_ips
            },
            LOG,
            exception,
            driver=self,
            active_backend_id=active_backend_id,
            host=self.host)

    @staticmethod
    def get_driver_options():
        return driver_opts

    def do_setup(self, context):
        """Setup and verify connection to IBM Storage."""

        self.proxy.setup(context)

    def ensure_export(self, context, volume):
        """Ensure an export."""

        return self.proxy.ensure_export(context, volume)

    def create_export(self, context, volume, connector):
        """Create an export."""

        return self.proxy.create_export(context, volume)

    def create_volume(self, volume):
        """Create a volume on the IBM Storage system."""

        return self.proxy.create_volume(volume)

    def delete_volume(self, volume):
        """Delete a volume on the IBM Storage system."""

        self.proxy.delete_volume(volume)

    def remove_export(self, context, volume):
        """Disconnect a volume from an attached instance."""

        return self.proxy.remove_export(context, volume)

    def initialize_connection(self, volume, connector):
        """Map the created volume."""

        conn_info = self.proxy.initialize_connection(volume, connector)
        fczm_utils.add_fc_zone(conn_info)
        return conn_info

    def terminate_connection(self, volume, connector, **kwargs):
        """Terminate a connection to a volume."""

        conn_info = self.proxy.terminate_connection(volume, connector)
        fczm_utils.remove_fc_zone(conn_info)
        return conn_info

    def create_volume_from_snapshot(self, volume, snapshot):
        """Create a volume from a snapshot."""

        return self.proxy.create_volume_from_snapshot(
            volume,
            snapshot)

    def create_snapshot(self, snapshot):
        """Create a snapshot."""

        return self.proxy.create_snapshot(snapshot)

    def delete_snapshot(self, snapshot):
        """Delete a snapshot."""

        return self.proxy.delete_snapshot(snapshot)

    def get_volume_stats(self, refresh=False):
        """Get volume stats."""

        return self.proxy.get_volume_stats(refresh)

    def create_cloned_volume(self, tgt_volume, src_volume):
        """Create Cloned Volume."""

        return self.proxy.create_cloned_volume(tgt_volume, src_volume)

    def extend_volume(self, volume, new_size):
        """Extend Created Volume."""

        self.proxy.extend_volume(volume, new_size)

    def migrate_volume(self, context, volume, host):
        """Migrate the volume to the specified host."""

        return self.proxy.migrate_volume(context, volume, host)

    def manage_existing(self, volume, existing_ref):
        """Brings an existing backend storage object to Cinder management."""

        return self.proxy.manage_volume(volume, existing_ref)

    def manage_existing_get_size(self, volume, existing_ref):
        """Return size of volume to be managed by manage_existing."""

        return self.proxy.manage_volume_get_size(volume, existing_ref)

    def unmanage(self, volume):
        """Removes the specified volume from Cinder management."""

        return self.proxy.unmanage_volume(volume)

    def freeze_backend(self, context):
        """Notify the backend that it's frozen. """

        return self.proxy.freeze_backend(context)

    def thaw_backend(self, context):
        """Notify the backend that it's unfrozen/thawed. """

        return self.proxy.thaw_backend(context)

    def failover_host(self, context, volumes, secondary_id=None, groups=None):
        """Failover a backend to a secondary replication target. """

        return self.proxy.failover_host(
            context, volumes, secondary_id, groups)

    def get_replication_status(self, context, volume):
        """Return replication status."""

        return self.proxy.get_replication_status(context, volume)

    def retype(self, ctxt, volume, new_type, diff, host):
        """Convert the volume to be of the new type."""

        return self.proxy.retype(ctxt, volume, new_type, diff, host)

    def revert_to_snapshot(self, ctxt, volume, snapshot):
        """Revert volume to snapshot."""
        return self.proxy.revert_to_snapshot(ctxt, volume, snapshot)

    def create_group(self, context, group):
        """Creates a group."""

        return self.proxy.create_group(context, group)

    def delete_group(self, context, group, volumes):
        """Deletes a group."""

        return self.proxy.delete_group(context, group, volumes)

    def create_group_snapshot(self, context, group_snapshot, snapshots):
        """Creates a group snapshot."""

        return self.proxy.create_group_snapshot(
            context, group_snapshot, snapshots)

    def delete_group_snapshot(self, context, group_snapshot, snapshots):
        """Deletes a group snapshot."""

        return self.proxy.delete_group_snapshot(
            context, group_snapshot, snapshots)

    def update_group(self, context, group, add_volumes, remove_volumes):
        """Adds or removes volume(s) to/from an existing group."""

        return self.proxy.update_group(
            context, group, add_volumes, remove_volumes)

    def create_group_from_src(
            self, context, group, volumes, group_snapshot, snapshots,
            source_cg=None, source_vols=None):
        """Creates a group from source."""

        return self.proxy.create_group_from_src(
            context, group, volumes, group_snapshot, snapshots,
            source_cg, source_vols)

    def enable_replication(self, context, group, volumes):
        """Enable replication."""

        return self.proxy.enable_replication(context, group, volumes)

    def disable_replication(self, context, group, volumes):
        """Disable replication."""

        return self.proxy.disable_replication(context, group, volumes)

    def failover_replication(self, context, group, volumes,
                             secondary_backend_id):
        """Failover replication."""

        return self.proxy.failover_replication(context, group, volumes,
                                               secondary_backend_id)

    def get_replication_error_status(self, context, groups):
        """Returns error info for replicated groups and its volumes."""

        return self.proxy.get_replication_error_status(context, groups)



#    (c) Copyright 2019 Brocade, a Broadcom Company
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from oslo_config import cfg
from oslo_log import log as logging

from cinder.volume import configuration

brcd_zone_opts = [
    cfg.StrOpt('fc_southbound_protocol',
               default='REST_HTTP',
               choices=('SSH', 'HTTP', 'HTTPS', 'REST_HTTP', 'REST_HTTPS'),
               help='South bound connector for the fabric.'),
    cfg.StrOpt('fc_fabric_address',
               default='',
               help='Management IP of fabric.'),
    cfg.StrOpt('fc_fabric_user',
               default='',
               help='Fabric user ID.'),
    cfg.StrOpt('fc_fabric_password',
               default='',
               help='Password for user.',
               secret=True),
    cfg.PortOpt('fc_fabric_port',
                default=22,
                help='Connecting port'),
    cfg.StrOpt('fc_fabric_ssh_cert_path',
               default='',
               help='Local SSH certificate Path.'),
    cfg.StrOpt('zoning_policy',
               default='initiator-target',
               help='Overridden zoning policy.'),
    cfg.BoolOpt('zone_activate',
                default=True,
                help='Overridden zoning activation state.'),
    cfg.StrOpt('zone_name_prefix',
               default='openstack',
               help='Overridden zone name prefix.'),
    cfg.StrOpt('fc_virtual_fabric_id',
               default=None,
               help='Virtual Fabric ID.')
]

CONF = cfg.CONF
CONF.register_opts(brcd_zone_opts, group='BRCD_FABRIC_EXAMPLE')
LOG = logging.getLogger(__name__)


def load_fabric_configurations(fabric_names):
    fabric_configs = {}
    for fabric_name in fabric_names:
        config = configuration.Configuration(brcd_zone_opts, fabric_name)
        LOG.debug("Loaded FC fabric config %(fabricname)s",
                  {'fabricname': fabric_name})
        fabric_configs[fabric_name] = config
    return fabric_configs



#    (c) Copyright 2014 Brocade Communications Systems Inc.
#    All Rights Reserved.
#
#    Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


class FCCommon(object):
    """Common interface for FC operations."""

    VERSION = "1.0"

    def __init__(self, **kwargs):
        pass

    def get_version(self):
        return self.VERSION



#    (c) Copyright 2014 Cisco Systems Inc.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


import random

from eventlet import greenthread
from oslo_concurrency import processutils
from oslo_log import log as logging
from oslo_utils import excutils
import six

from cinder import exception
from cinder.i18n import _
from cinder import ssh_utils
from cinder import utils
from cinder.zonemanager.drivers.cisco import cisco_fabric_opts as fabric_opts
from cinder.zonemanager.drivers.cisco import exception as c_exception
import cinder.zonemanager.drivers.cisco.fc_zone_constants as zone_constant
from cinder.zonemanager import fc_san_lookup_service as fc_service
from cinder.zonemanager import utils as zm_utils

LOG = logging.getLogger(__name__)


class CiscoFCSanLookupService(fc_service.FCSanLookupService):
    """The SAN lookup service that talks to Cisco switches.

    Version History:
        1.0.0 - Initial version

    """

    VERSION = "1.0.0"

    def __init__(self, **kwargs):
        """Initializing the client."""
        super(CiscoFCSanLookupService, self).__init__(**kwargs)
        self.configuration = kwargs.get('configuration', None)
        self.create_configuration()

        self.switch_user = ""
        self.switch_port = ""
        self.switch_pwd = ""
        self.switch_ip = ""
        self.sshpool = {}

    def create_configuration(self):
        """Configuration specific to SAN context values."""
        config = self.configuration

        fabric_names = [x.strip() for x in config.fc_fabric_names.split(',')]
        LOG.debug('Fabric Names: %s', fabric_names)

        # There can be more than one SAN in the network and we need to
        # get credentials for each for SAN context lookup later.
        # Cisco Zonesets require VSANs
        if fabric_names:
            self.fabric_configs = fabric_opts.load_fabric_configurations(
                fabric_names)

    def get_device_mapping_from_network(self,
                                        initiator_wwn_list,
                                        target_wwn_list):
        """Provides the initiator/target map for available SAN contexts.

        Looks up fcns database of each fc SAN configured to find logged in
        devices and returns a map of initiator and target port WWNs for each
        fabric.

        :param initiator_wwn_list: List of initiator port WWN
        :param target_wwn_list: List of target port WWN
        :returns: List -- device wwn map in following format

        .. code-block:: python

            {
                <Fabric name>: {
                    'initiator_port_wwn_list':
                    ('200000051e55a100', '200000051e55a121'..)
                    'target_port_wwn_list':
                    ('100000051e55a100', '100000051e55a121'..)
                }
            }

        :raises Exception: when connection to fabric is failed
        """
        device_map = {}
        formatted_target_list = []
        formatted_initiator_list = []
        fabric_map = {}
        fabric_names = self.configuration.fc_fabric_names

        if not fabric_names:
            raise exception.InvalidParameterValue(
                err=_("Missing Fibre Channel SAN configuration "
                      "param - fc_fabric_names"))

        fabrics = [x.strip() for x in fabric_names.split(',')]

        LOG.debug("FC Fabric List: %s", fabrics)
        if fabrics:
            for t in target_wwn_list:
                formatted_target_list.append(zm_utils.get_formatted_wwn(t))

            for i in initiator_wwn_list:
                formatted_initiator_list.append(zm_utils.get_formatted_wwn(i))

            for fabric_name in fabrics:
                self.switch_ip = self.fabric_configs[fabric_name].safe_get(
                    'cisco_fc_fabric_address')
                self.switch_user = self.fabric_configs[fabric_name].safe_get(
                    'cisco_fc_fabric_user')
                self.switch_pwd = self.fabric_configs[fabric_name].safe_get(
                    'cisco_fc_fabric_password')
                self.switch_port = self.fabric_configs[fabric_name].safe_get(
                    'cisco_fc_fabric_port')
                zoning_vsan = self.fabric_configs[fabric_name].safe_get(
                    'cisco_zoning_vsan')

                # Get name server data from fabric and find the targets
                # logged in
                nsinfo = ''
                LOG.debug("show fcns database for vsan %s", zoning_vsan)
                nsinfo = self.get_nameserver_info(zoning_vsan)

                self.cleanup()

                LOG.debug("Lookup service:fcnsdatabase-%s", nsinfo)
                LOG.debug("Lookup service:initiator list from caller-%s",
                          formatted_initiator_list)
                LOG.debug("Lookup service:target list from caller-%s",
                          formatted_target_list)
                visible_targets = [x for x in nsinfo
                                   if x in formatted_target_list]
                visible_initiators = [x for x in nsinfo
                                      if x in formatted_initiator_list]

                if visible_targets:
                    LOG.debug("Filtered targets is: %s", visible_targets)
                    # getting rid of the : before returning
                    for idx, elem in enumerate(visible_targets):
                        elem = str(elem).replace(':', '')
                        visible_targets[idx] = elem
                else:
                    LOG.debug("No targets are in the fcns database"
                              " for vsan %s", zoning_vsan)

                if visible_initiators:
                    # getting rid of the : before returning ~sk
                    for idx, elem in enumerate(visible_initiators):
                        elem = str(elem).replace(':', '')
                        visible_initiators[idx] = elem
                else:
                    LOG.debug("No initiators are in the fcns database"
                              " for vsan %s", zoning_vsan)

                fabric_map = {'initiator_port_wwn_list': visible_initiators,
                              'target_port_wwn_list': visible_targets
                              }
                device_map[fabric_name] = fabric_map
        LOG.debug("Device map for SAN context: %s", device_map)
        return device_map

    def get_nameserver_info(self, fabric_vsan):
        """Get fcns database info from fabric.

        This method will return the connected node port wwn list(local
        and remote) for the given switch fabric
        """
        cli_output = None
        nsinfo_list = []
        try:
            cmd = ([zone_constant.FCNS_SHOW, fabric_vsan, ' | no-more'])
            cli_output = self._get_switch_info(cmd)
        except exception.FCSanLookupServiceException:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed collecting show fcns database for fabric")
        if cli_output:
            nsinfo_list = self._parse_ns_output(cli_output)

        LOG.debug("Connector returning fcns info-%s", nsinfo_list)
        return nsinfo_list

    def _get_switch_info(self, cmd_list):
        stdout, stderr, sw_data = None, None, None
        try:
            stdout, stderr = self._run_ssh(cmd_list, True, 1)
            LOG.debug("CLI output from ssh - output: %s", stdout)
            if (stdout):
                sw_data = stdout.splitlines()
            return sw_data
        except processutils.ProcessExecutionError as e:
            msg = _("Error while getting data via ssh: (command=%(cmd)s "
                    "error=%(err)s).") % {'cmd': cmd_list,
                                          'err': six.text_type(e)}
            LOG.error(msg)
            raise c_exception.CiscoZoningCliException(reason=msg)

    def _parse_ns_output(self, switch_data):
        """Parses name server data.

        Parses nameserver raw data and adds the device port wwns to the list

        :returns: list of device port wwn from ns info
        """
        nsinfo_list = []
        for line in switch_data:
            if not(" N " in line):
                continue
            linesplit = line.split()
            if len(linesplit) > 2:
                node_port_wwn = linesplit[2]
                nsinfo_list.append(node_port_wwn)
            else:
                msg = _("Malformed fcns output string: %s") % line
                LOG.error(msg)
                raise exception.InvalidParameterValue(err=msg)
        return nsinfo_list

    def _run_ssh(self, cmd_list, check_exit_code=True, attempts=1):

        command = ' '.join(cmd_list)

        if self.sshpool.get(self.switch_ip) is None:
            self.sshpool[self.switch_ip] = ssh_utils.SSHPool(self.switch_ip,
                                                             self.switch_port,
                                                             None,
                                                             self.switch_user,
                                                             self.switch_pwd,
                                                             min_size=1,
                                                             max_size=5)

        last_exception = None
        try:
            with self.sshpool.get(self.switch_ip).item() as ssh:
                while attempts > 0:
                    attempts -= 1
                    try:
                        return processutils.ssh_execute(
                            ssh,
                            command,
                            check_exit_code=check_exit_code)
                    except Exception as e:
                        msg = _("Exception: %s") % six.text_type(e)
                        LOG.error(msg)
                        last_exception = e
                        greenthread.sleep(random.randint(20, 500) / 100.0)
                try:
                    raise processutils.ProcessExecutionError(
                        exit_code=last_exception.exit_code,
                        stdout=last_exception.stdout,
                        stderr=last_exception.stderr,
                        cmd=last_exception.cmd)
                except AttributeError:
                    raise processutils.ProcessExecutionError(
                        exit_code=-1,
                        stdout="",
                        stderr="Error running SSH command",
                        cmd=command)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error("Error running SSH command: %s", command)

    def _ssh_execute(self, cmd_list, check_exit_code=True, attempts=1):
        """Execute cli with status update.

        Executes CLI commands where status return is expected.

        cmd_list is a list of commands, where each command is itself
        a list of parameters.  We use utils.check_ssh_injection to check each
        command, but then join then with " ; " to form a single command.
        """

        # Check that each command is secure
        for cmd in cmd_list:
            utils.check_ssh_injection(cmd)

        # Combine into a single command.
        command = ' ; '.join(map(lambda x: ' '.join(x), cmd_list))

        if self.sshpool.get(self.switch_ip) is None:
            self.sshpool[self.switch_ip] = ssh_utils.SSHPool(self.switch_ip,
                                                             self.switch_port,
                                                             None,
                                                             self.switch_user,
                                                             self.switch_pwd,
                                                             min_size=1,
                                                             max_size=5)

        stdin, stdout, stderr = None, None, None
        LOG.debug("Executing command via ssh: %s", command)
        last_exception = None
        try:
            with self.sshpool.get(self.switch_ip).item() as ssh:
                while attempts > 0:
                    attempts -= 1
                    try:
                        stdin, stdout, stderr = ssh.exec_command(command)
                        greenthread.sleep(random.randint(20, 500) / 100.0)
                        channel = stdout.channel
                        exit_status = channel.recv_exit_status()
                        LOG.debug("Exit Status from ssh:%s", exit_status)
                        # exit_status == -1 if no exit code was returned
                        if exit_status != -1:
                            LOG.debug('Result was %s', exit_status)
                            if check_exit_code and exit_status != 0:
                                raise processutils.ProcessExecutionError(
                                    exit_code=exit_status,
                                    stdout=stdout,
                                    stderr=stderr,
                                    cmd=command)
                            else:
                                return True
                        else:
                            return True
                    except Exception as e:
                        msg = _("Exception: %s") % six.text_type(e)
                        LOG.error(msg)
                        last_exception = e
                        greenthread.sleep(random.randint(20, 500) / 100.0)
                LOG.debug("Handling error case after SSH:%s", last_exception)
                try:
                    raise processutils.ProcessExecutionError(
                        exit_code=last_exception.exit_code,
                        stdout=last_exception.stdout,
                        stderr=last_exception.stderr,
                        cmd=last_exception.cmd)
                except AttributeError:
                    raise processutils.ProcessExecutionError(
                        exit_code=-1,
                        stdout="",
                        stderr="Error running SSH command",
                        cmd=command)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                msg = (_("Error executing command via ssh: %s") %
                       six.text_type(e))
                LOG.error(msg)
        finally:
            if stdin:
                stdin.flush()
                stdin.close()
            if stdout:
                stdout.close()
            if stderr:
                stderr.close()

    def cleanup(self):
        self.sshpool = {}



#
# Copyright (c) 2016 NEC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Drivers for M-Series Storage."""

from cinder import interface
from cinder.volume import driver
from cinder.volume.drivers.nec import volume_common
from cinder.volume.drivers.nec import volume_helper
from cinder.zonemanager import utils as fczm_utils


@interface.volumedriver
class MStorageISCSIDriver(volume_helper.MStorageDSVDriver,
                          driver.ISCSIDriver):
    """M-Series Storage Snapshot iSCSI Driver.

    .. code-block:: none

      Version history:

        1.8.1 - First open source driver version.
        1.8.2 - Code refactoring.
        1.9.1 - Support optimal path for non-disruptive backup.
        1.9.2 - Support manage/unmanage and manage/unmanage snapshot.
                Delete an unused configuration
                parameter (ldset_controller_node_name).
                Fixed bug #1705001: driver fails to start.
        1.10.1 - Support automatic configuration of SAN access control.
                 Fixed bug #1753375: SAN access remains permitted on the
                 source node.
        1.10.2 - Delete max volumes per pool limit.
        1.10.3 - Add faster clone status check.
                 Fixed bug #1777385: driver removed access permission from
                 the destination node after live-migraion.
                 Fixed bug #1778669: LUNs of detached volumes are never reused.
        1.11.1 - Add support python 3.
                 Add support for multi-attach.
                 Add support of more than 4 iSCSI portals for a node.
                 Add support to revert a volume to a snapshot.
                 Add support storage assist retype and fixed bug #1838955:
                 a volume in NEC Storage was left undeleted when the volume
                 was retyped to another storage.
    """

    VERSION = '1.11.1'
    CI_WIKI_NAME = 'NEC_Cinder_CI'

    def __init__(self, *args, **kwargs):
        super(MStorageISCSIDriver, self).__init__(*args, **kwargs)
        self._set_config(self.configuration, self.host,
                         self.__class__.__name__)

    @staticmethod
    def get_driver_options():
        return volume_common.mstorage_opts

    def ensure_export(self, context, volume):
        pass

    def get_volume_stats(self, refresh=False):
        return self.iscsi_get_volume_stats(refresh)

    def initialize_connection(self, volume, connector):
        return self.iscsi_initialize_connection(volume, connector)

    def terminate_connection(self, volume, connector, **kwargs):
        return self.iscsi_terminate_connection(volume, connector)

    def initialize_connection_snapshot(self, snapshot, connector, **kwargs):
        return self.iscsi_initialize_connection_snapshot(snapshot,
                                                         connector,
                                                         **kwargs)

    def terminate_connection_snapshot(self, snapshot, connector, **kwargs):
        return self.iscsi_terminate_connection_snapshot(snapshot,
                                                        connector,
                                                        **kwargs)


@interface.volumedriver
class MStorageFCDriver(volume_helper.MStorageDSVDriver,
                       driver.FibreChannelDriver):
    """M-Series Storage Snapshot FC Driver.

    .. code-block:: none

      Version history:

        1.8.1 - First open source driver version.
        1.8.2 - Code refactoring.
        1.9.1 - Support optimal path for non-disruptive backup.
        1.9.2 - Support manage/unmanage and manage/unmanage snapshot.
                Delete an unused configuration
                parameter (ldset_controller_node_name).
                Fixed bug #1705001: driver fails to start.
        1.10.1 - Support automatic configuration of SAN access control.
                 Fixed bug #1753375: SAN access remains permitted on the
                 source node.
        1.10.2 - Delete max volumes per pool limit.
        1.10.3 - Add faster clone status check.
                 Fixed bug #1777385: driver removed access permission from
                 the destination node after live-migraion.
                 Fixed bug #1778669: LUNs of detached volumes are never reused.
        1.11.1 - Add support python 3.
                 Add support for multi-attach.
                 Add support of more than 4 iSCSI portals for a node.
                 Add support to revert a volume to a snapshot.
                 Add support storage assist retype and fixed bug #1838955:
                 a volume in NEC Storage was left undeleted when the volume
                 was retyped to another storage.
    """

    VERSION = '1.11.1'
    CI_WIKI_NAME = 'NEC_Cinder_CI'

    def __init__(self, *args, **kwargs):
        super(MStorageFCDriver, self).__init__(*args, **kwargs)
        self._set_config(self.configuration, self.host,
                         self.__class__.__name__)

    @staticmethod
    def get_driver_options():
        return volume_common.mstorage_opts

    def ensure_export(self, context, volume):
        pass

    def get_volume_stats(self, refresh=False):
        return self.fc_get_volume_stats(refresh)

    def initialize_connection(self, volume, connector):
        conn_info = self.fc_initialize_connection(volume, connector)
        fczm_utils.add_fc_zone(conn_info)
        return conn_info

    def terminate_connection(self, volume, connector, **kwargs):
        conn_info = self.fc_terminate_connection(volume, connector)
        fczm_utils.remove_fc_zone(conn_info)
        return conn_info

    def initialize_connection_snapshot(self, snapshot, connector, **kwargs):
        return self.fc_initialize_connection_snapshot(snapshot,
                                                      connector,
                                                      **kwargs)

    def terminate_connection_snapshot(self, snapshot, connector, **kwargs):
        return self.fc_terminate_connection_snapshot(snapshot,
                                                     connector,
                                                     **kwargs)



# Copyright 2018 Inspur Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Volume driver test for Inspur AS13000."""

import json
import random
import time
from unittest import mock

import ddt
import eventlet
import requests

from cinder import context
from cinder import exception
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.volume import configuration
from cinder.volume.drivers.inspur.as13000 import as13000_driver
from cinder.volume import volume_utils


test_config = configuration.Configuration(None)
test_config.san_ip = 'some_ip'
test_config.san_api_port = 'as13000_api_port'
test_config.san_login = 'username'
test_config.san_password = 'password'
test_config.as13000_ipsan_pools = ['fakepool']
test_config.as13000_meta_pool = 'meta_pool'
test_config.use_chap_auth = True
test_config.chap_username = 'fakeuser'
test_config.chap_password = 'fakepass'


class FakeResponse(object):
    def __init__(self, status, output):
        self.status_code = status
        self.text = 'return message'
        self._json = output

    def json(self):
        return self._json

    def close(self):
        pass


@ddt.ddt
class RestAPIExecutorTestCase(test.TestCase):
    def setUp(self):
        self.rest_api = as13000_driver.RestAPIExecutor(
            test_config.san_ip,
            test_config.san_api_port,
            test_config.san_login,
            test_config.san_password)
        super(RestAPIExecutorTestCase, self).setUp()

    def test_login(self):
        mock__login = self.mock_object(self.rest_api, '_login',
                                       mock.Mock(return_value='fake_token'))
        self.rest_api.login()
        mock__login.assert_called_once()
        self.assertEqual('fake_token', self.rest_api._token)

    def test__login(self):
        response = {'token': 'fake_token', 'expireTime': '7200', 'type': 0}
        mock_sra = self.mock_object(self.rest_api, 'send_rest_api',
                                    mock.Mock(return_value=response))

        result = self.rest_api._login()

        self.assertEqual('fake_token', result)

        login_params = {'name': test_config.san_login,
                        'password': test_config.san_password}
        mock_sra.assert_called_once_with(method='security/token',
                                         params=login_params,
                                         request_type='post')

    def test_send_rest_api(self):
        expected = {'value': 'abc'}
        mock_sa = self.mock_object(self.rest_api, 'send_api',
                                   mock.Mock(return_value=expected))
        result = self.rest_api.send_rest_api(
            method='fake_method',
            params='fake_params',
            request_type='fake_type')
        self.assertEqual(expected, result)
        mock_sa.assert_called_once_with(
            'fake_method',
            'fake_params',
            'fake_type')

    def test_send_rest_api_retry(self):
        expected = {'value': 'abc'}
        mock_sa = self.mock_object(
            self.rest_api,
            'send_api',
            mock.Mock(side_effect=(exception.VolumeDriverException, expected)))
        mock_login = self.mock_object(self.rest_api, 'login', mock.Mock())
        result = self.rest_api.send_rest_api(
            method='fake_method',
            params='fake_params',
            request_type='fake_type'
        )
        self.assertEqual(expected, result)

        mock_sa.assert_called_with(
            'fake_method',
            'fake_params',
            'fake_type')
        mock_login.assert_called_once()

    def test_send_rest_api_3times_fail(self):
        mock_sa = self.mock_object(
            self.rest_api, 'send_api', mock.Mock(
                side_effect=(exception.VolumeDriverException)))
        mock_login = self.mock_object(self.rest_api, 'login', mock.Mock())
        self.assertRaises(
            exception.VolumeDriverException,
            self.rest_api.send_rest_api,
            method='fake_method',
            params='fake_params',
            request_type='fake_type')
        mock_sa.assert_called_with('fake_method',
                                   'fake_params',
                                   'fake_type')
        mock_login.assert_called()

    def test_send_rest_api_backend_error_fail(self):
        side_effect = exception.VolumeBackendAPIException('fake_err_msg')
        mock_sa = self.mock_object(self.rest_api,
                                   'send_api',
                                   mock.Mock(side_effect=side_effect))
        mock_login = self.mock_object(self.rest_api, 'login')

        self.assertRaises(exception.VolumeBackendAPIException,
                          self.rest_api.send_rest_api,
                          method='fake_method',
                          params='fake_params',
                          request_type='fake_type')
        mock_sa.assert_called_with('fake_method',
                                   'fake_params',
                                   'fake_type')
        mock_login.assert_not_called()

    @ddt.data(
        {'method': 'fake_method', 'request_type': 'post', 'params':
            {'fake_param': 'fake_value'}},
        {'method': 'fake_method', 'request_type': 'get', 'params':
            {'fake_param': 'fake_value'}},
        {'method': 'fake_method', 'request_type': 'delete', 'params':
            {'fake_param': 'fake_value'}},
        {'method': 'fake_method', 'request_type': 'put', 'params':
            {'fake_param': 'fake_value'}}, )
    @ddt.unpack
    def test_send_api(self, method, params, request_type):
        self.rest_api._token = 'fake_token'
        if request_type in ('post', 'delete', 'put'):
            fake_output = {'code': 0, 'message': 'success'}
        elif request_type == 'get':
            fake_output = {'code': 0, 'data': 'fake_date'}
        mock_request = self.mock_object(
            requests, request_type, mock.Mock(
                return_value=FakeResponse(
                    200, fake_output)))
        self.rest_api.send_api(
            method,
            params=params,
            request_type=request_type)
        mock_request.assert_called_once_with(
            'http://%s:%s/rest/%s' %
            (test_config.san_ip,
             test_config.san_api_port,
             method),
            data=json.dumps(params),
            headers={'X-Auth-Token': 'fake_token'})

    @ddt.data({'method': r'security/token',
               'params': {'name': test_config.san_login,
                          'password': test_config.san_password},
               'request_type': 'post'},
              {'method': r'security/token',
               'params': None,
               'request_type': 'delete'})
    @ddt.unpack
    def test_send_api_access_success(self, method, params, request_type):
        if request_type == 'post':
            fake_value = {'code': 0, 'data': {
                'token': 'fake_token',
                'expireTime': '7200',
                'type': 0}}
            mock_requests = self.mock_object(
                requests, 'post', mock.Mock(
                    return_value=FakeResponse(
                        200, fake_value)))
            result = self.rest_api.send_api(method, params, request_type)
            self.assertEqual(fake_value['data'], result)
            mock_requests.assert_called_once_with(
                'http://%s:%s/rest/%s' %
                (test_config.san_ip,
                 test_config.san_api_port,
                 method),
                data=json.dumps(params),
                headers=None)
        if request_type == 'delete':
            fake_value = {'code': 0, 'message': 'Success!'}
            self.rest_api._token = 'fake_token'
            mock_requests = self.mock_object(
                requests, 'delete', mock.Mock(
                    return_value=FakeResponse(
                        200, fake_value)))
            self.rest_api.send_api(method, params, request_type)
            mock_requests.assert_called_once_with(
                'http://%s:%s/rest/%s' %
                (test_config.san_ip,
                 test_config.san_api_port,
                 method),
                data=None,
                headers={'X-Auth-Token': 'fake_token'})

    def test_send_api_wrong_access_fail(self):
        req_params = {'method': r'security/token',
                      'params': {'name': test_config.san_login,
                                 'password': 'fake_password'},
                      'request_type': 'post'}
        fake_value = {'message': ' User name or password error.', 'code': 400}
        mock_request = self.mock_object(
            requests, 'post', mock.Mock(
                return_value=FakeResponse(
                    200, fake_value)))
        self.assertRaises(
            exception.VolumeBackendAPIException,
            self.rest_api.send_api,
            method=req_params['method'],
            params=req_params['params'],
            request_type=req_params['request_type'])
        mock_request.assert_called_once_with(
            'http://%s:%s/rest/%s' %
            (test_config.san_ip,
             test_config.san_api_port,
             req_params['method']),
            data=json.dumps(
                req_params['params']),
            headers=None)

    def test_send_api_token_overtime_fail(self):
        self.rest_api._token = 'fake_token'
        fake_value = {'method': 'fake_url',
                      'params': 'fake_params',
                      'reuest_type': 'post'}
        fake_out_put = {'message': 'Unauthorized access!', 'code': 301}
        mock_requests = self.mock_object(
            requests, 'post', mock.Mock(
                return_value=FakeResponse(
                    200, fake_out_put)))
        self.assertRaises(exception.VolumeDriverException,
                          self.rest_api.send_api,
                          method='fake_url',
                          params='fake_params',
                          request_type='post')
        mock_requests.assert_called_once_with(
            'http://%s:%s/rest/%s' %
            (test_config.san_ip,
             test_config.san_api_port,
             fake_value['method']),
            data=json.dumps('fake_params'),
            headers={
                'X-Auth-Token': 'fake_token'})

    def test_send_api_fail(self):
        self.rest_api._token = 'fake_token'
        fake_output = {'code': 999, 'message': 'fake_message'}
        mock_request = self.mock_object(
            requests, 'post', mock.Mock(
                return_value=FakeResponse(
                    200, fake_output)))
        self.assertRaises(
            exception.VolumeBackendAPIException,
            self.rest_api.send_api,
            method='fake_method',
            params='fake_params',
            request_type='post')
        mock_request.assert_called_once_with(
            'http://%s:%s/rest/%s' %
            (test_config.san_ip,
             test_config.san_api_port,
             'fake_method'),
            data=json.dumps('fake_params'),
            headers={'X-Auth-Token': 'fake_token'}
        )


@ddt.ddt
class AS13000DriverTestCase(test.TestCase):
    def __init__(self, *args, **kwds):
        super(AS13000DriverTestCase, self).__init__(*args, **kwds)
        self._ctxt = context.get_admin_context()
        self.configuration = test_config

    def setUp(self):
        self.rest_api = as13000_driver.RestAPIExecutor(
            test_config.san_ip,
            test_config.san_api_port,
            test_config.san_login,
            test_config.san_password)
        self.as13000_san = as13000_driver.AS13000Driver(
            configuration=self.configuration)
        super(AS13000DriverTestCase, self).setUp()

    @ddt.data(None, 'pool1')
    def test_do_setup(self, meta_pool):
        mock_login = self.mock_object(as13000_driver.RestAPIExecutor,
                                      'login', mock.Mock())
        fake_nodes = [{'healthStatus': 1, 'ip': 'fakeip1'},
                      {'healthStatus': 1, 'ip': 'fakeip2'},
                      {'healthStatus': 1, 'ip': 'fakeip3'}]
        mock_gcs = self.mock_object(self.as13000_san,
                                    '_get_cluster_status',
                                    mock.Mock(return_value=fake_nodes))
        fake_pools = {
            'pool1': {'name': 'pool1', 'type': '1'},
            'pool2': {'name': 'pool2', 'type': 2}
        }
        mock_gpi = self.mock_object(self.as13000_san,
                                    '_get_pools_info',
                                    mock.Mock(return_value=fake_pools))
        mock_cp = self.mock_object(self.as13000_san,
                                   '_check_pools',
                                   mock.Mock())
        mock_cmp = self.mock_object(self.as13000_san,
                                    '_check_meta_pool',
                                    mock.Mock())

        self.as13000_san.meta_pool = meta_pool
        self.as13000_san.pools = ['pool1', 'pool2']
        self.as13000_san.do_setup(self._ctxt)

        mock_login.assert_called_once()
        mock_gcs.assert_called_once()
        if meta_pool:
            mock_gpi.assert_called_with(['pool1', 'pool2', 'pool1'])
        else:
            mock_gpi.assert_called_with(['pool1', 'pool2'])
            self.assertEqual('pool1', self.as13000_san.meta_pool)
        mock_cp.assert_called_once()
        mock_cmp.assert_called_once()

    def test_check_for_setup_error(self):
        mock_sg = self.mock_object(configuration.Configuration, 'safe_get',
                                   mock.Mock(return_value='fake_config'))
        self.as13000_san.nodes = [{'fakenode': 'fake_name'}]
        self.as13000_san.check_for_setup_error()
        mock_sg.assert_called()

    def test_check_for_setup_error_no_healthy_node_fail(self):
        mock_sg = self.mock_object(configuration.Configuration, 'safe_get',
                                   mock.Mock(return_value='fake_config'))
        self.as13000_san.nodes = []
        self.assertRaises(exception.VolumeDriverException,
                          self.as13000_san.check_for_setup_error)
        mock_sg.assert_called()

    def test_check_for_setup_error_no_config_fail(self):
        mock_sg = self.mock_object(configuration.Configuration, 'safe_get',
                                   mock.Mock(return_value=None))
        self.as13000_san.nodes = []
        self.assertRaises(exception.InvalidConfigurationValue,
                          self.as13000_san.check_for_setup_error)
        mock_sg.assert_called()

    def test__check_pools(self):
        fake_pools_info = {
            'pool1': {'name': 'pool1', 'type': '1'},
            'pool2': {'name': 'pool2', 'type': 1}
        }
        self.as13000_san.pools = ['pool1']
        self.as13000_san.pools_info = fake_pools_info
        self.as13000_san._check_pools()

    def test__check_pools_fail(self):
        fake_pools_info = {
            'pool1': {'name': 'pool1', 'type': '1'},
            'pool2': {'name': 'pool2', 'type': 1}
        }
        self.as13000_san.pools = ['pool0, pool1']
        self.as13000_san.pools_info = fake_pools_info
        self.assertRaises(exception.InvalidInput,
                          self.as13000_san._check_pools)

    def test__check_meta_pool(self):
        fake_pools_info = {
            'pool1': {'name': 'pool1', 'type': 2},
            'pool2': {'name': 'pool2', 'type': 1}
        }
        self.as13000_san.meta_pool = 'pool2'
        self.as13000_san.pools_info = fake_pools_info
        self.as13000_san._check_meta_pool()

    @ddt.data(None, 'pool0', 'pool1')
    def test__check_meta_pool_failed(self, meta_pool):
        fake_pools_info = {
            'pool1': {'name': 'pool1', 'type': 2},
            'pool2': {'name': 'pool2', 'type': 1}
        }

        self.as13000_san.meta_pool = meta_pool
        self.as13000_san.pools_info = fake_pools_info
        self.assertRaises(exception.InvalidInput,
                          self.as13000_san._check_meta_pool)

    @mock.patch.object(as13000_driver.RestAPIExecutor,
                       'send_rest_api')
    def test_create_volume(self, mock_rest):
        volume = fake_volume.fake_volume_obj(self._ctxt, host='H@B#P')
        self.as13000_san.pools_info = {'P': {'name': 'P', 'type': 1}}
        self.as13000_san.meta_pool = 'meta_pool'
        self.as13000_san.create_volume(volume)

        mock_rest.assert_called_once_with(
            method='block/lvm',
            params={
                "name": volume.name.replace('-', '_'),
                "capacity": volume.size * 1024,
                "dataPool": 'P',
                "dataPoolType": 1,
                "metaPool": 'meta_pool'
            },
            request_type='post')

    @ddt.data(1, 2)
    def test_create_volume_from_snapshot(self, size):
        volume = fake_volume.fake_volume_obj(self._ctxt, size=size)
        volume2 = fake_volume.fake_volume_obj(self._ctxt)
        snapshot = fake_snapshot.fake_snapshot_obj(self._ctxt, volume=volume2)

        mock_eh = self.mock_object(volume_utils,
                                   'extract_host',
                                   mock.Mock(return_value='fake_pool'))
        _tnd_mock = mock.Mock(side_effect=('source_volume',
                                           'dest_volume',
                                           'snapshot'))
        mock_tnd = self.mock_object(self.as13000_san,
                                    '_trans_name_down',
                                    _tnd_mock)
        mock_lock_op = self.mock_object(self.as13000_san,
                                        '_snapshot_lock_op',
                                        mock.Mock())
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock())
        mock_fv = self.mock_object(self.as13000_san,
                                   '_filling_volume',
                                   mock.Mock())
        mock_wvf = self.mock_object(self.as13000_san,
                                    '_wait_volume_filled',
                                    mock.Mock())
        mock_ev = self.mock_object(self.as13000_san, 'extend_volume',
                                   mock.Mock())

        self.as13000_san.create_volume_from_snapshot(volume, snapshot)

        lock_op_calls = [
            mock.call('lock', 'source_volume', 'snapshot', 'fake_pool'),
            mock.call('unlock', 'source_volume', 'snapshot', 'fake_pool')
        ]
        mock_lock_op.assert_has_calls(lock_op_calls)
        mock_fv.assert_called_once_with('dest_volume', 'fake_pool')
        mock_wvf.assert_called_once_with('dest_volume', 'fake_pool')

        mock_eh.assert_called()
        mock_tnd.assert_called()
        params = {
            'originalLvm': 'source_volume',
            'originalPool': 'fake_pool',
            'originalSnap': 'snapshot',
            'name': 'dest_volume',
            'pool': 'fake_pool'}
        mock_rest.assert_called_once_with(method='snapshot/volume/cloneLvm',
                                          params=params,
                                          request_type='post')
        if size == 2:
            mock_ev.assert_called_once_with(volume, size)

    def test_create_volume_from_snapshot_fail(self):
        volume = fake_volume.fake_volume_obj(self._ctxt)
        snapshot = fake_snapshot.fake_snapshot_obj(self._ctxt, volume_size=10)
        self.assertRaises(
            exception.InvalidInput,
            self.as13000_san.create_volume_from_snapshot, volume, snapshot)

    @ddt.data(1, 2)
    def test_create_cloned_volume(self, size):
        volume = fake_volume.fake_volume_obj(self._ctxt, size=size)
        volume_src = fake_volume.fake_volume_obj(self._ctxt)
        mock_eh = self.mock_object(volume_utils,
                                   'extract_host',
                                   mock.Mock(return_value='fake_pool'))
        mock_tnd = self.mock_object(
            self.as13000_san, '_trans_name_down', mock.Mock(
                side_effect=('fake_name1', 'fake_name2')))
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock())
        mock_ev = self.mock_object(self.as13000_san,
                                   'extend_volume',
                                   mock.Mock())
        self.as13000_san.create_cloned_volume(volume, volume_src)
        mock_eh.assert_called()
        mock_tnd.assert_called()
        method = 'block/lvm/clone'
        params = {
            'srcVolumeName': 'fake_name2',
            'srcPoolName': 'fake_pool',
            'destVolumeName': 'fake_name1',
            'destPoolName': 'fake_pool'}
        request_type = 'post'
        mock_rest.assert_called_once_with(
            method=method, params=params, request_type=request_type)
        if size == 2:
            mock_ev.assert_called_once_with(volume, size)

    def test_create_clone_volume_fail(self):
        volume = fake_volume.fake_volume_obj(self._ctxt)
        volume_source = fake_volume.fake_volume_obj(self._ctxt, size=2)
        self.assertRaises(
            exception.InvalidInput,
            self.as13000_san.create_cloned_volume, volume, volume_source)

    def test_extend_volume(self):
        volume = fake_volume.fake_volume_obj(self._ctxt)
        mock_tnd = self.mock_object(
            self.as13000_san, '_trans_name_down', mock.Mock(
                return_value='fake_name'))
        mock_cv = self.mock_object(self.as13000_san,
                                   '_check_volume',
                                   mock.Mock(return_value=True))
        mock_eh = self.mock_object(volume_utils,
                                   'extract_host',
                                   mock.Mock(return_value='fake_pool'))
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock())
        self.as13000_san.extend_volume(volume, 10)
        mock_tnd.assert_called_once_with(volume.name)
        mock_cv.assert_called_once_with(volume)
        mock_eh.assert_called_once_with(volume.host, level='pool')
        method = 'block/lvm'
        request_type = 'put'
        params = {'pool': 'fake_pool',
                  'name': 'fake_name',
                  'newCapacity': 10240}
        mock_rest.assert_called_once_with(method=method,
                                          request_type=request_type,
                                          params=params)

    def test_extend_volume_fail(self):
        volume = fake_volume.fake_volume_obj(self._ctxt)
        mock_tnd = self.mock_object(
            self.as13000_san, '_trans_name_down', mock.Mock(
                return_value='fake_name'))
        mock_cv = self.mock_object(self.as13000_san,
                                   '_check_volume',
                                   mock.Mock(return_value=False))
        self.assertRaises(exception.VolumeDriverException,
                          self.as13000_san.extend_volume, volume, 10)
        mock_tnd.assert_called_once_with(volume.name)
        mock_cv.assert_called_once_with(volume)

    @ddt.data(True, False)
    def test_delete_volume(self, volume_exist):
        volume = fake_volume.fake_volume_obj(self._ctxt)
        mock_eh = self.mock_object(volume_utils,
                                   'extract_host',
                                   mock.Mock(return_value='fake_pool'))
        mock_tnd = self.mock_object(
            self.as13000_san, '_trans_name_down', mock.Mock(
                return_value='fake_name'))
        mock_cv = self.mock_object(self.as13000_san,
                                   '_check_volume',
                                   mock.Mock(return_value=volume_exist))
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock())
        self.as13000_san.delete_volume(volume)
        mock_tnd.assert_called_once_with(volume.name)
        mock_cv.assert_called_once_with(volume)

        if volume_exist:
            mock_eh.assert_called_once_with(volume.host, level='pool')

            method = 'block/lvm?pool=%s&lvm=%s' % ('fake_pool', 'fake_name')
            request_type = 'delete'
            mock_rest.assert_called_once_with(method=method,
                                              request_type=request_type)

    def test_create_snapshot(self):
        volume = fake_volume.fake_volume_obj(self._ctxt)
        snapshot = fake_snapshot.fake_snapshot_obj(self._ctxt, volume=volume)
        mock_eh = self.mock_object(volume_utils,
                                   'extract_host',
                                   mock.Mock(return_value='fake_pool'))
        mock_cv = self.mock_object(self.as13000_san,
                                   '_check_volume',
                                   mock.Mock(return_value=True))
        mock_tnd = self.mock_object(
            self.as13000_san, '_trans_name_down', mock.Mock(
                side_effect=('fake_name', 'fake_snap')))
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock())
        self.as13000_san.create_snapshot(snapshot)

        mock_eh.assert_called_once_with(volume.host, level='pool')
        mock_tnd.assert_called()
        mock_cv.assert_called_once_with(snapshot.volume)
        method = 'snapshot/volume'
        params = {'snapName': 'fake_snap',
                  'volumeName': 'fake_name',
                  'poolName': 'fake_pool',
                  'snapType': 'r'}
        request_type = 'post'
        mock_rest.assert_called_once_with(method=method,
                                          params=params,
                                          request_type=request_type)

    def test_create_snapshot_fail(self):
        volume = fake_volume.fake_volume_obj(self._ctxt)
        snapshot = fake_snapshot.fake_snapshot_obj(self._ctxt, volume=volume)
        mock_cv = self.mock_object(self.as13000_san,
                                   '_check_volume',
                                   mock.Mock(return_value=False))
        self.assertRaises(exception.VolumeDriverException,
                          self.as13000_san.create_snapshot, snapshot)
        mock_cv.assert_called_once_with(snapshot.volume)

    def test_delete_snapshot(self):
        volume = fake_volume.fake_volume_obj(self._ctxt)
        snapshot = fake_snapshot.fake_snapshot_obj(self._ctxt, volume=volume)
        mock_eh = self.mock_object(volume_utils,
                                   'extract_host',
                                   mock.Mock(return_value='fake_pool'))
        mock_cv = self.mock_object(self.as13000_san,
                                   '_check_volume',
                                   mock.Mock(return_value=True))
        mock_tnd = self.mock_object(
            self.as13000_san, '_trans_name_down', mock.Mock(
                side_effect=('fake_name', 'fake_snap')))
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock())
        self.as13000_san.delete_snapshot(snapshot)

        mock_eh.assert_called_once_with(volume.host, level='pool')
        mock_tnd.assert_called()
        mock_cv.assert_called_once_with(snapshot.volume)

        method = ('snapshot/volume?snapName=%s&volumeName=%s&poolName=%s'
                  % ('fake_snap', 'fake_name', 'fake_pool'))
        request_type = 'delete'
        mock_rest.assert_called_once_with(method=method,
                                          request_type=request_type)

    def test_delete_snapshot_fail(self):
        volume = fake_volume.fake_volume_obj(self._ctxt)
        snapshot = fake_snapshot.fake_snapshot_obj(self._ctxt, volume=volume)
        mock_cv = self.mock_object(self.as13000_san,
                                   '_check_volume',
                                   mock.Mock(return_value=False))
        self.assertRaises(exception.VolumeDriverException,
                          self.as13000_san.delete_snapshot, snapshot)
        mock_cv.assert_called_once_with(snapshot.volume)

    @mock.patch('time.time')
    @ddt.data(2000, 1000)
    def test__update_volume_stats(self, time_token, mock_time):
        mock_time.return_value = 5000
        self.as13000_san.VENDOR = 'INSPUR'
        self.as13000_san.VERSION = 'V1.3.1'
        self.as13000_san.PROTOCOL = 'iSCSI'
        mock_sg = self.mock_object(configuration.Configuration, 'safe_get',
                                   mock.Mock(return_value='fake_backend_name'))
        fake_pool_backend = [{'pool_name': 'fake_pool'},
                             {'pool_name': 'fake_pool1'}]
        self.as13000_san.pools = ['fake_pool']
        mock_gps = self.mock_object(self.as13000_san, '_get_pools_stats',
                                    mock.Mock(return_value=fake_pool_backend))
        self.as13000_san._stats = None
        self.as13000_san._token_time = time_token
        self.as13000_san.token_available_time = 3600
        mock_login = self.mock_object(as13000_driver.RestAPIExecutor,
                                      'login')

        self.as13000_san._update_volume_stats()
        backend_data = {'driver_version': 'V1.3.1',
                        'pools': fake_pool_backend,
                        'storage_protocol': 'iSCSI',
                        'vendor_name': 'INSPUR',
                        'volume_backend_name': 'fake_backend_name'}

        self.assertEqual(backend_data, self.as13000_san._stats)
        mock_sg.assert_called_once_with('volume_backend_name')
        mock_gps.assert_called_once()
        if (time.time() - time_token) > 3600:
            mock_login.assert_called_once()
        else:
            mock_login.assert_not_called()

    @ddt.data((4, u'127.0.0.1', '3260'),
              (6, u'FF01::101', '3260'))
    @ddt.unpack
    def test__build_target_portal(self, version, ip, port):
        portal = self.as13000_san._build_target_portal(ip, port)
        if version == 4:
            self.assertEqual(portal, '127.0.0.1:3260')
        else:
            self.assertEqual(portal, '[FF01::101]:3260')

    @ddt.data((True, True, True),
              (True, True, False),
              (False, True, True),
              (False, True, False),
              (False, False, True),
              (False, False, False),
              (True, False, True),
              (True, False, False))
    @ddt.unpack
    def test_initialize_connection(self, host_exist, multipath, chap_enabled):
        volume = fake_volume.fake_volume_obj(self._ctxt)
        connector = {'multipath': multipath,
                     'ip': 'fake_ip',
                     'host': 'fake_host'}
        self.as13000_san.configuration.use_chap_auth = chap_enabled
        fakenode = [{'name': 'fake_name1', 'ip': 'node_ip1'},
                    {'name': 'fake_name2', 'ip': 'node_ip2'},
                    {'name': 'fake_name3', 'ip': 'node_ip3'}]
        self.as13000_san.nodes = fakenode
        if multipath:
            mock_gtfc = self.mock_object(
                self.as13000_san,
                '_get_target_from_conn',
                mock.Mock(return_value=(host_exist,
                                        'target_name',
                                        ['fake_name1', 'fake_name2'])))
        else:
            mock_gtfc = self.mock_object(
                self.as13000_san,
                '_get_target_from_conn',
                mock.Mock(return_value=(host_exist,
                                        'target_name',
                                        ['fake_name1'])))

        mock_altt = self.mock_object(self.as13000_san,
                                     '_add_lun_to_target',
                                     mock.Mock())
        mock_ct = self.mock_object(self.as13000_san,
                                   '_create_target',
                                   mock.Mock())
        mock_ahtt = self.mock_object(self.as13000_san,
                                     '_add_host_to_target',
                                     mock.Mock())
        mock_actt = self.mock_object(self.as13000_san,
                                     '_add_chap_to_target',
                                     mock.Mock())
        mock_gli = self.mock_object(self.as13000_san,
                                    '_get_lun_id',
                                    mock.Mock(return_value='1'))
        mock_rr = self.mock_object(random, 'randint',
                                   mock.Mock(return_value='12345678'))
        mock_btp = self.mock_object(self.as13000_san,
                                    '_build_target_portal',
                                    mock.Mock(side_effect=['node_ip1:3260',
                                                           'node_ip2:3260',
                                                           'node_ip3:3260']))

        connect_info = self.as13000_san.initialize_connection(
            volume, connector)

        expect_conn_data = {
            'target_discovered': True,
            'volume_id': volume.id,
        }
        if host_exist:
            if multipath:
                expect_conn_data.update({
                    'target_portals': ['node_ip1:3260', 'node_ip2:3260'],
                    'target_luns': [1] * 2,
                    'target_iqns': ['target_name'] * 2
                })
            else:
                expect_conn_data.update({
                    'target_portal': 'node_ip1:3260',
                    'target_lun': 1,
                    'target_iqn': 'target_name'
                })
        else:
            target_name = 'target.inspur.fake_host-12345678'
            if multipath:
                expect_conn_data.update({
                    'target_portals': ['node_ip1:3260',
                                       'node_ip2:3260',
                                       'node_ip3:3260'],
                    'target_luns': [1] * 3,
                    'target_iqns': [target_name] * 3
                })
            else:
                expect_conn_data.update({
                    'target_portal': 'node_ip1:3260',
                    'target_lun': 1,
                    'target_iqn': target_name
                })

        if chap_enabled:
            expect_conn_data['auth_method'] = 'CHAP'
            expect_conn_data['auth_username'] = 'fakeuser'
            expect_conn_data['auth_password'] = 'fakepass'

        expect_datas = {
            'driver_volume_type': 'iscsi',
            'data': expect_conn_data
        }

        self.assertEqual(expect_datas, connect_info)
        mock_gtfc.assert_called_once_with('fake_ip')
        mock_altt.assert_called_once()
        if not host_exist:
            mock_ct.assert_called_once()
            mock_ahtt.assert_called_once()
            mock_rr.assert_called_once()
        if chap_enabled:
            mock_actt.assert_called_once()

        mock_gli.assert_called_once()
        mock_btp.assert_called()

    @ddt.data(True, False)
    def test_terminate_connection(self, delete_target):
        volume = fake_volume.fake_volume_obj(self._ctxt, host='fakehost')
        connector = {'multipath': False,
                     'ip': 'fake_ip',
                     'host': 'fake_host'}
        mock_tnd = self.mock_object(self.as13000_san, '_trans_name_down',
                                    mock.Mock(return_value='fake_volume'))
        fake_target_list = [{'hostIp': ['fake_ip'],
                             'name': 'target_name',
                             'lun': [
                                 {'lvm': 'fake_volume', 'lunID': 'fake_id'}]}]
        mock_gtl = self.mock_object(self.as13000_san, '_get_target_list',
                                    mock.Mock(return_value=fake_target_list))
        mock_dlft = self.mock_object(self.as13000_san,
                                     '_delete_lun_from_target',
                                     mock.Mock())
        if delete_target:
            mock_gll = self.mock_object(self.as13000_san, '_get_lun_list',
                                        mock.Mock(return_value=[]))
        else:
            mock_gll = self.mock_object(self.as13000_san, '_get_lun_list',
                                        mock.Mock(return_value=[1, 2]))
        mock_dt = self.mock_object(self.as13000_san, '_delete_target',
                                   mock.Mock())
        self.as13000_san.terminate_connection(volume, connector)
        mock_tnd.assert_called_once_with(volume.name)
        mock_gtl.assert_called_once()
        mock_dlft.assert_called_once_with(lun_id='fake_id',
                                          target_name='target_name')
        mock_gll.assert_called_once_with('target_name')
        if delete_target:
            mock_dt.assert_called_once_with('target_name')
        else:
            mock_dt.assert_not_called()

    @ddt.data(True, False)
    def test_terminate_connection_force(self, delete_target):
        volume = fake_volume.fake_volume_obj(self._ctxt, host='fakehost')
        connector = {}
        mock_tnd = self.mock_object(self.as13000_san, '_trans_name_down',
                                    mock.Mock(return_value='fake_volume'))
        fake_target_list = [{'hostIp': ['fake_hostIp'],
                             'name':'target_name',
                             'lun':[{'lvm': 'fake_volume',
                                     'lunID': 'fake_id'}]}]
        mock_gtl = self.mock_object(self.as13000_san, '_get_target_list',
                                    mock.Mock(return_value=fake_target_list))
        mock_dlft = self.mock_object(self.as13000_san,
                                     '_delete_lun_from_target',
                                     mock.Mock())
        if delete_target:
            mock_gll = self.mock_object(self.as13000_san, '_get_lun_list',
                                        mock.Mock(return_value=[]))
        else:
            mock_gll = self.mock_object(self.as13000_san, '_get_lun_list',
                                        mock.Mock(return_value=[1, 2]))
        mock_dt = self.mock_object(self.as13000_san, '_delete_target',
                                   mock.Mock())

        self.as13000_san.terminate_connection(volume, connector)

        mock_tnd.assert_called_once_with(volume.name)
        mock_gtl.assert_called_once()
        mock_dlft.assert_called_once_with(lun_id='fake_id',
                                          target_name='target_name')
        mock_gll.assert_called_once_with('target_name')
        if delete_target:
            mock_dt.assert_called_once_with('target_name')
        else:
            mock_dt.assert_not_called()

    @mock.patch.object(as13000_driver.RestAPIExecutor,
                       'send_rest_api')
    def test__get_pools_info(self, mock_rest):
        fake_pools_data = [{'name': 'pool1', 'type': 1},
                           {'name': 'pool2', 'type': 2}]
        mock_rest.return_value = fake_pools_data

        # get a partial of pools
        result_pools_info = self.as13000_san._get_pools_info(['pool1'])
        self.assertEqual(result_pools_info,
                         {'pool1': {'name': 'pool1', 'type': 1}})

        # get both exist pools
        result_pools_info = self.as13000_san._get_pools_info(['pool1',
                                                              'pool2'])
        self.assertEqual(result_pools_info,
                         {'pool1': {'name': 'pool1', 'type': 1},
                          'pool2': {'name': 'pool2', 'type': 2}})

        # get pools not exist
        result_pools_info = self.as13000_san._get_pools_info(['pool1',
                                                              'pool2',
                                                              'pool3'])
        self.assertEqual(result_pools_info,
                         {'pool1': {'name': 'pool1', 'type': 1},
                          'pool2': {'name': 'pool2', 'type': 2}})

    def test__get_pools_stats(self):
        pool_date = [{'ID': 'fake_id',
                      'name': 'fake_name',
                      'totalCapacity': '2t',
                      'usedCapacity': '300g'}]
        self.as13000_san.pools = ['fake_name']
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock(return_value=pool_date))
        mock_uc = self.mock_object(self.as13000_san, '_unit_convert',
                                   mock.Mock(side_effect=(2000, 300)))
        pool_info = {
            'pool_name': 'fake_name',
            'total_capacity_gb': 2000,
            'free_capacity_gb': 1700,
            'thin_provisioning_support': True,
            'thick_provisioning_support': False,
        }

        result_pools = self.as13000_san._get_pools_stats()

        expect_pools = [pool_info]
        self.assertEqual(expect_pools, result_pools)
        mock_rest.assert_called_once_with(method='block/pool?type=2',
                                          request_type='get')
        mock_uc.assert_called()

    @ddt.data('fake_ip3', 'fake_ip5')
    def test__get_target_from_conn(self, host_ip):
        target_list = [
            {
                'hostIp': ['fake_ip1', 'fake_ip2'],
                'name':'fake_target_1',
                'node':['fake_node1', 'fake_node2']
            },
            {
                'hostIp': ['fake_ip3', 'fake_ip4'],
                'name': 'fake_target_2',
                'node': ['fake_node4', 'fake_node3']
            }
        ]
        mock_gtl = self.mock_object(self.as13000_san,
                                    '_get_target_list',
                                    mock.Mock(return_value=target_list))

        host_exist, target_name, node = (
            self.as13000_san._get_target_from_conn(host_ip))

        if host_ip == 'fake_ip3':
            self.assertEqual((True, 'fake_target_2',
                              ['fake_node4', 'fake_node3']),
                             (host_exist, target_name, node))
        else:
            self.assertEqual((False, None, None),
                             (host_exist, target_name, node))
        mock_gtl.assert_called_once()

    def test__get_target_list(self):
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock(return_value='fake_date'))
        method = 'block/target/detail'
        request_type = 'get'
        result = self.as13000_san._get_target_list()
        self.assertEqual('fake_date', result)
        mock_rest.assert_called_once_with(method=method,
                                          request_type=request_type)

    def test__create_target(self):
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock())
        target_name = 'fake_name'
        target_node = 'fake_node'
        method = 'block/target'
        params = {'name': target_name, 'nodeName': target_node}
        request_type = 'post'
        self.as13000_san._create_target(target_name, target_node)
        mock_rest.assert_called_once_with(method=method,
                                          params=params,
                                          request_type=request_type)

    def test__delete_target(self):
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock())
        target_name = 'fake_name'
        method = 'block/target?name=%s' % target_name
        request_type = 'delete'
        self.as13000_san._delete_target(target_name)
        mock_rest.assert_called_once_with(method=method,
                                          request_type=request_type)

    def test__add_chap_to_target(self):
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock())
        target_name = 'fake_name'
        chap_username = 'fake_user'
        chap_password = 'fake_pass'
        self.as13000_san._add_chap_to_target(target_name,
                                             chap_username,
                                             chap_password)

        method = 'block/chap/bond'
        params = {'target': target_name,
                  'user': chap_username,
                  'password': chap_password}
        request_type = 'post'
        mock_rest.assert_called_once_with(method=method,
                                          params=params,
                                          request_type=request_type)

    def test__add_host_to_target(self):
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock())
        target_name = 'fake_name'
        host_ip = 'fake_ip'
        method = 'block/host'
        params = {'name': target_name, 'hostIp': host_ip}
        request_type = 'post'
        self.as13000_san._add_host_to_target(host_ip, target_name)
        mock_rest.assert_called_once_with(method=method,
                                          params=params,
                                          request_type=request_type)

    def test__add_lun_to_target(self):
        volume = fake_volume.fake_volume_obj(self._ctxt, host='fakehost')
        mock_eh = self.mock_object(volume_utils,
                                   'extract_host',
                                   mock.Mock(return_value='fake_pool'))
        mock_tnd = self.mock_object(self.as13000_san,
                                    '_trans_name_down',
                                    mock.Mock(return_value='fake_name'))
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock())

        target_name = 'fake_target'
        self.as13000_san._add_lun_to_target(target_name, volume)
        method = 'block/lun'
        params = {'name': target_name,
                  'pool': 'fake_pool',
                  'lvm': 'fake_name'}
        request_type = 'post'
        mock_eh.assert_called_once_with(volume.host, level='pool')
        mock_tnd.assert_called_once_with(volume.name)
        mock_rest.assert_called_once_with(method=method,
                                          params=params,
                                          request_type=request_type)

    def test__add_lun_to_target_retry_3times(self):
        volume = fake_volume.fake_volume_obj(self._ctxt, host='fakehost')
        mock_eh = self.mock_object(volume_utils,
                                   'extract_host',
                                   mock.Mock(return_value='fake_pool'))
        mock_tnd = self.mock_object(self.as13000_san,
                                    '_trans_name_down',
                                    mock.Mock(return_value='fake_name'))
        mock_rest = self.mock_object(
            as13000_driver.RestAPIExecutor,
            'send_rest_api',
            mock.Mock(side_effect=(exception.VolumeDriverException,
                                   mock.MagicMock())))

        target_name = 'fake_target'
        self.as13000_san._add_lun_to_target(target_name, volume)
        method = 'block/lun'
        params = {'name': target_name,
                  'pool': 'fake_pool',
                  'lvm': 'fake_name'}
        request_type = 'post'
        mock_eh.assert_called_with(volume.host, level='pool')
        mock_tnd.assert_called_with(volume.name)
        mock_rest.assert_called_with(method=method,
                                     params=params,
                                     request_type=request_type)

    def test__add_lun_to_target_fail(self):
        volume = fake_volume.fake_volume_obj(self._ctxt, host='fakehost')
        mock_eh = self.mock_object(volume_utils,
                                   'extract_host',
                                   mock.Mock(return_value='fake_pool'))
        mock_tnd = self.mock_object(self.as13000_san,
                                    '_trans_name_down',
                                    mock.Mock(return_value='fake_name'))
        mock_rest = self.mock_object(
            as13000_driver.RestAPIExecutor,
            'send_rest_api',
            mock.Mock(side_effect=exception.VolumeDriverException))

        target_name = 'fake_target'
        self.assertRaises(exception.VolumeDriverException,
                          self.as13000_san._add_lun_to_target,
                          target_name=target_name,
                          volume=volume)
        method = 'block/lun'
        params = {'name': target_name,
                  'pool': 'fake_pool',
                  'lvm': 'fake_name'}
        request_type = 'post'
        mock_eh.assert_called_with(volume.host, level='pool')
        mock_tnd.assert_called_with(volume.name)
        mock_rest.assert_called_with(method=method,
                                     params=params,
                                     request_type=request_type)

    def test__delete_lun_from_target(self):
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock())
        target_name = 'fake_target'
        lun_id = 'fake_id'
        self.as13000_san._delete_lun_from_target(target_name, lun_id)
        method = 'block/lun?name=%s&id=%s&force=1' % (target_name, lun_id)
        request_type = 'delete'
        mock_rest.assert_called_once_with(method=method,
                                          request_type=request_type)

    @ddt.data('lock', 'unlock')
    def test__snapshot_lock_op(self, operation):
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock())
        vol_name = 'fake_volume'
        snap_name = 'fake_snapshot'
        pool_name = "fake_pool"
        self.as13000_san._snapshot_lock_op(operation,
                                           vol_name,
                                           snap_name,
                                           pool_name)

        method = 'snapshot/volume/' + operation
        request_type = 'post'
        params = {'snapName': snap_name,
                  'volumeName': vol_name,
                  'poolName': pool_name}
        mock_rest.assert_called_once_with(method=method,
                                          params=params,
                                          request_type=request_type)

    def test__filling_volume(self):
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock())
        vol_name = 'fake_volume'
        pool_name = 'fake_pool'
        self.as13000_san._filling_volume(vol_name, pool_name)
        params = {'pool': 'fake_pool', 'name': 'fake_volume'}
        mock_rest.assert_called_once_with(method='block/lvm/filling',
                                          params=params,
                                          request_type='post')

    def test__wait_volume_filled(self):
        # Need to mock sleep as it is called by @utils.retry
        self.mock_object(time, 'sleep')

        expected = [{'name': 'fake_v1', 'lvmType': 1}]
        mock_gv = self.mock_object(self.as13000_san, '_get_volumes',
                                   mock.Mock(return_value=expected))
        self.as13000_san._wait_volume_filled('fake_v1', 'fake_pool')
        mock_gv.assert_called_with('fake_pool')

    def test__wait_volume_filled_failed(self):
        # Need to mock sleep as it is called by @utils.retry
        self.mock_object(time, 'sleep')

        expected = [{'name': 'fake_v1', 'lvmType': 2}]
        mock_gv = self.mock_object(self.as13000_san, '_get_volumes',
                                   mock.Mock(return_value=expected))
        self.assertRaises(exception.VolumeDriverException,
                          self.as13000_san._wait_volume_filled,
                          'fake_v1',
                          'fake_pool')
        mock_gv.assert_called_with('fake_pool')

    def test__get_lun_list(self):
        target_name = 'fake_name'
        lun_list = ['lun_1', 'lun_2']
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock(return_value=lun_list))
        lun_result = self.as13000_san._get_lun_list(target_name)
        self.assertEqual(lun_list, lun_result)
        method = 'block/lun?name=%s' % target_name
        request_type = 'get'
        mock_rest.assert_called_once_with(method=method,
                                          request_type=request_type)

    @ddt.data(True, False)
    def test__check_volume(self, exist):
        volume = fake_volume.fake_volume_obj(self._ctxt, host='fakehost')
        mock_eh = self.mock_object(volume_utils,
                                   'extract_host',
                                   mock.Mock(return_value='fake_pool'))
        mock_tnd = self.mock_object(self.as13000_san,
                                    '_trans_name_down',
                                    mock.Mock(return_value='fake_name'))
        mock_el = self.mock_object(eventlet, 'sleep',
                                   mock.Mock(return_value=None))
        if exist:
            mock_gv = self.mock_object(self.as13000_san, '_get_volumes',
                                       mock.Mock(return_value=[
                                           {'name': 'fake_name'},
                                           {'name': 'fake_name2'}]))
        else:
            mock_gv = self.mock_object(self.as13000_san, '_get_volumes',
                                       mock.Mock(return_value=[
                                           {'name': 'fake_name2'},
                                           {'name': 'fake_name3'}]))
        expect = self.as13000_san._check_volume(volume)
        self.assertEqual(exist, expect)
        mock_eh.assert_called_once_with(volume.host, 'pool')
        mock_tnd.assert_called_once_with(volume.name)
        if exist:
            mock_gv.assert_called_once_with('fake_pool')
        else:
            mock_gv.assert_called()
            mock_el.assert_called()

    def test__get_volumes(self):
        volumes = [{'name': 'fake_name1'},
                   {'name': 'fake_name2'},
                   {'name': 'fake_name3'}]
        pool = 'fake_pool'
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock(return_value=volumes))
        result = self.as13000_san._get_volumes(pool)
        method = 'block/lvm?pool=%s' % pool
        request_type = 'get'
        self.assertEqual(volumes, result)
        mock_rest.assert_called_once_with(method=method,
                                          request_type=request_type)

    def test__get_cluster_status(self):
        method = 'cluster/node'
        request_type = 'get'
        cluster = 'fake_cluster'
        mock_rest = self.mock_object(as13000_driver.RestAPIExecutor,
                                     'send_rest_api',
                                     mock.Mock(return_value=cluster))
        result = self.as13000_san._get_cluster_status()
        self.assertEqual(cluster, result)
        mock_rest.assert_called_once_with(method=method,
                                          request_type=request_type)

    @ddt.data(True, False)
    def test__get_lun_id(self, lun_exist):
        volume = fake_volume.fake_volume_obj(self._ctxt, host='fakehost')
        if lun_exist:
            lun_list = [{'id': '01', 'mappingLvm': r'fake_pool/fake_volume1'},
                        {'id': '02', 'mappingLvm': r'fake_pool/fake_volume2'}]
        else:
            lun_list = [{'id': '01', 'mappingLvm': r'fake_pool/fake_volume1'},
                        {'id': '02', 'mappingLvm': r'fake_pool/fake_volume0'}]

        mock_eh = self.mock_object(volume_utils,
                                   'extract_host',
                                   mock.Mock(return_value='fake_pool'))
        mock_tnd = self.mock_object(self.as13000_san,
                                    '_trans_name_down',
                                    mock.Mock(return_value='fake_volume2'))
        mock_gll = self.mock_object(self.as13000_san, '_get_lun_list',
                                    mock.Mock(return_value=lun_list))

        lun_id = self.as13000_san._get_lun_id(volume, 'fake_target')
        if lun_exist:
            self.assertEqual('02', lun_id)
        else:
            self.assertIsNone(lun_id)

        mock_eh.assert_called_once_with(volume.host, level='pool')
        mock_tnd.assert_called_once_with(volume.name)
        mock_gll.assert_called_once_with('fake_target')

    def test__trans_name_down(self):
        fake_name = 'test-abcd-1234_1234-234'
        expect = 'test_abcd_1234_1234_234'
        result = self.as13000_san._trans_name_down(fake_name)
        self.assertEqual(expect, result)

    @ddt.data('5000000000', '5000000k', '5000mb', '50G', '5TB', '5PB', '5EB')
    def test__unit_convert(self, capacity):
        trans = {'5000000000': '%.0f' % (float(5000000000) / (1024 ** 3)),
                 '5000000k': '%.0f' % (float(5000000) / (1024 ** 2)),
                 '5000mb': '%.0f' % (float(5000) / 1024),
                 '50G': '%.0f' % float(50),
                 '5TB': '%.0f' % (float(5) * 1024),
                 '5PB': '%.0f' % (float(5) * (1024 ** 2)),
                 '5EB': '%.0f' % (float(5) * (1024 ** 3))}
        expect = float(trans[capacity])
        result = self.as13000_san._unit_convert(capacity)
        self.assertEqual(expect, result)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt

from cinder.api import microversions as mv
from cinder.api.v3 import group_snapshots
from cinder import exception
from cinder.group import api as group_api
from cinder.objects import fields
from cinder.policies import group_snapshots as group_snap_policies
from cinder.policies import groups as group_policies
from cinder.tests.unit.api import fakes as fake_api
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit.policies import base
from cinder.tests.unit import utils as test_utils
from cinder.volume import manager as volume_manager


@ddt.ddt
class GroupSnapshotsPolicyTest(base.BasePolicyTest):
    authorized_users = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
    ]
    unauthorized_users = [
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    authorized_members = authorized_users
    unauthorized_members = unauthorized_users

    authorized_readers = authorized_users
    unauthorized_readers = unauthorized_users

    sysadmins = [
        'legacy_admin',
        'project_admin',
        'system_admin',
    ]
    non_sysadmins = [
        'legacy_owner',
        'project_member',
        'project_reader',
        'project_foo',
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    # Basic policy test is without enforcing scope (which cinder doesn't
    # yet support) and deprecated rules enabled.
    def setUp(self, enforce_scope=False, enforce_new_defaults=False,
              *args, **kwargs):
        super().setUp(enforce_scope, enforce_new_defaults, *args, **kwargs)
        self.controller = group_snapshots.GroupSnapshotsController()
        self.manager = volume_manager.VolumeManager()
        self.api_path = '/v3/%s/group_snapshots' % (self.project_id)
        self.api_version = mv.GROUP_GROUPSNAPSHOT_PROJECT_ID

    def _create_group_snapshot(
            self, snap_status=fields.GroupSnapshotStatus.AVAILABLE):
        volume_type = test_utils.create_volume_type(self.project_admin_context,
                                                    name="test")

        group = test_utils.create_group(self.project_admin_context,
                                        status=fields.GroupStatus.AVAILABLE,
                                        group_type_id=fake.GROUP_TYPE_ID,
                                        volume_type_ids=[volume_type.id])

        test_utils.create_volume(self.project_member_context,
                                 group_id=group.id,
                                 testcase_instance=self,
                                 volume_type_id=volume_type.id)

        return test_utils.create_group_snapshot(
            self.project_admin_context,
            group_id=group.id,
            status=snap_status,
            group_type_id=group.group_type_id)

    def _create_group_snap_array(self):
        group = test_utils.create_group(self.project_admin_context,
                                        status=fields.GroupStatus.AVAILABLE,
                                        group_type_id=fake.GROUP_TYPE_ID,
                                        volume_type_ids=[fake.VOLUME_TYPE_ID])

        test_utils.create_volume(self.project_member_context,
                                 group_id=group.id,
                                 volume_type_id=fake.VOLUME_TYPE_ID)
        g_snapshots_array = [
            test_utils.create_group_snapshot(
                self.project_admin_context,
                group_id=group.id,
                group_type_id=group.group_type_id) for _ in range(3)]

        return g_snapshots_array

    @ddt.data(*base.all_users)
    def test_create_group_snapshot_policy(self, user_id):
        """Test create a group snapshot."""
        volume_type = test_utils.create_volume_type(self.project_admin_context,
                                                    name='test')

        group = test_utils.create_group(self.project_admin_context,
                                        status=fields.GroupStatus.AVAILABLE,
                                        group_type_id=fake.GROUP_TYPE_ID,
                                        volume_type_ids=[volume_type.id])

        test_utils.create_volume(self.project_member_context,
                                 group_id=group.id,
                                 testcase_instance=self,
                                 volume_type_id=volume_type.id)

        rule_name = group_snap_policies.CREATE_POLICY
        version = mv.GROUP_GROUPSNAPSHOT_PROJECT_ID
        url = self.api_path
        req = fake_api.HTTPRequest.blank(url, version=version)
        req.method = 'POST'
        body = {
            "group_snapshot": {
                "name": "my_group_snapshot",
                "description": "My group snapshot",
                "group_id": group.id,
            }
        }
        unauthorized_exceptions = [exception.GroupNotFound]

        # Relax the group:get policy in order to get past that check.
        self.policy.set_rules({group_policies.GET_POLICY: ""},
                              overwrite=False)

        self.common_policy_check(user_id,
                                 self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.controller.create,
                                 req,
                                 body=body)
        group.destroy()

    @ddt.data(*base.all_users)
    def test_update_group_snapshot_policy(self, user_id):
        # This call is not implemented in the Block Storage API v3
        # so we need to test group_snap_policies.UPDATE_POLICY directly
        # against the group API
        group_snapshot = self._create_group_snapshot()
        api = group_api.API()
        ctxt = self.create_context(user_id)
        if user_id in self.authorized_members:
            api.update_group_snapshot(ctxt, group_snapshot, {})
        elif user_id in self.unauthorized_members:
            self.assertRaises(exception.PolicyNotAuthorized,
                              api.update_group_snapshot,
                              ctxt,
                              group_snapshot,
                              {})
        else:
            self.fail(f'{user_id} not in authorized or unauthorized members')

    @ddt.data(*base.all_users)
    def test_delete_group_snapshot_policy(self, user_id):
        """Delete group snapshot."""
        # Redirect the RPC call directly to the volume manager.
        rule_name = group_snap_policies.DELETE_POLICY
        group_snapshot = self._create_group_snapshot()
        url = '%s/%s' % (self.api_path, group_snapshot.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'DELETE'
        unauthorized_exceptions = [exception.GroupSnapshotNotFound,
                                   exception.InvalidGroupSnapshot]

        # Relax the GET_POLICY in order to get past that check.
        self.policy.set_rules({group_snap_policies.GET_POLICY: ""},
                              overwrite=False)

        self.common_policy_check(user_id,
                                 self.authorized_members,
                                 self.unauthorized_members,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.controller.delete,
                                 req,
                                 id=group_snapshot.id)

    @ddt.data(*base.all_users)
    def test_get_all_group_snapshot_policy(self, user_id):
        """List group snapshots."""
        self._create_group_snap_array()
        rule_name = group_snap_policies.GET_ALL_POLICY
        url = '%s/detail' % (self.api_path)
        version = mv.GROUP_SNAPSHOTS
        req = fake_api.HTTPRequest.blank(url, version=version)
        unauthorized_exceptions = []

        # NOTE: we intentionally don't use the un/authorized_readers
        # lists in this function because get-all doesn't have a target
        # to authorize against
        #
        # legacy: any logged in user can list all group snapshots
        # (project-specific filtering happens later)
        authorized_users = [user_id]
        unauthorized_users = []
        # ... unless deprecated rules are not allowed, then you
        # must have the 'reader' role to read
        if self.enforce_new_defaults:
            context = self.create_context(user_id)
            if 'reader' not in context.roles:
                authorized_users = []
                unauthorized_users = [user_id]
        self.common_policy_check(user_id,
                                 authorized_users,
                                 unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.controller.detail,
                                 req)

    @ddt.data(*base.all_users)
    def test_get_group_snapshot_policy(self, user_id):
        """Show group snapshot."""

        group_snapshot = self._create_group_snapshot()
        rule_name = group_snap_policies.GET_POLICY
        url = '%s/%s' % (self.api_path, group_snapshot.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        unauthorized_exceptions = [exception.GroupSnapshotNotFound]
        self.common_policy_check(user_id,
                                 self.authorized_readers,
                                 self.unauthorized_readers,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.controller.show,
                                 req,
                                 id=group_snapshot.id)

    @ddt.data(*base.all_users)
    @mock.patch('cinder.api.v3.views.group_snapshots.ViewBuilder.detail')
    @mock.patch('cinder.group.api.API.get_group_snapshot')
    def test_group_snapshot_project_attribute_policy(self, user_id,
                                                     mock_api,
                                                     mock_view):
        """Test show group snapshot with project attributes."""

        # FIXME: kind of fragile, but I'm beginning to like this approach
        def mock_view_detail(request, group_snapshot):
            context = request.environ['cinder.context']
            context.authorize(
                group_snap_policies.GROUP_SNAPSHOT_ATTRIBUTES_POLICY)

        group_snapshot = self._create_group_snapshot()
        group_snapshot_id = group_snapshot.id
        mock_api.return_value = group_snapshot
        mock_view.side_effect = mock_view_detail

        rule_name = group_snap_policies.GROUP_SNAPSHOT_ATTRIBUTES_POLICY
        url = '%s/%s' % (self.api_path, group_snapshot_id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)

        unauthorized_exceptions = [exception.GroupSnapshotNotFound]
        self.common_policy_check(user_id,
                                 self.sysadmins,
                                 self.non_sysadmins,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.controller.show,
                                 req,
                                 id=group_snapshot_id)


class GroupSnapshotsPolicySecureRbacTest(GroupSnapshotsPolicyTest):
    authorized_users = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
    ]
    unauthorized_users = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    authorized_members = authorized_users
    unauthorized_members = unauthorized_users

    authorized_readers = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
    ]
    unauthorized_readers = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    def setUp(self, *args, **kwargs):
        # Test secure RBAC by disabling deprecated policy rules (scope
        # is still not enabled).
        super().setUp(enforce_scope=False, enforce_new_defaults=True,
                      *args, **kwargs)



#    (c) Copyright 2014 Cisco Systems Inc.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""Unit tests for Cisco fc zone client cli."""

import time
from unittest import mock

from oslo_concurrency import processutils

from cinder.tests.unit import test
from cinder.zonemanager.drivers.cisco \
    import cisco_fc_zone_client_cli as cli
from cinder.zonemanager.drivers.cisco import exception as c_exception
import cinder.zonemanager.drivers.cisco.fc_zone_constants as ZoneConstant

nsshow = '20:1a:00:05:1e:e8:e3:29'
switch_data = ['VSAN 303\n',
               '----------------------------------------------------------\n',
               'FCID     TYPE  PWWN           (VENDOR)    FC4-TYPE:FEATURE\n',
               '----------------------------------------------------------\n',
               '0x030001 N     20:1a:00:05:1e:e8:e3:29 (Cisco)        ipfc\n',
               '0x030101 NL    10:00:00:00:77:99:60:2c (Interphase)\n',
               '0x030200    NL    10:00:00:49:c9:28:c7:01\n']

cfgactv = ['zoneset name OpenStack_Cfg vsan 303\n',
           'zone name openstack50060b0000c26604201900051ee8e329 vsan 303\n',
           'pwwn 50:06:0b:00:00:c2:66:04\n',
           'pwwn 20:19:00:05:1e:e8:e3:29\n']

active_zoneset = {
    'zones': {
        'openstack50060b0000c26604201900051ee8e329':
        ['50:06:0b:00:00:c2:66:04', '20:19:00:05:1e:e8:e3:29']},
    'active_zone_config': 'OpenStack_Cfg'}

zoning_status_data_basic = [
    'VSAN: 303 default-zone: deny distribute: active only Interop: default\n',
    '    mode: basic merge-control: allow\n',
    '    session: none\n',
    '    hard-zoning: enabled broadcast: unsupported\n',
    '    smart-zoning: disabled\n',
    '    rscn-format: fabric-address\n',
    'Default zone:\n',
    '    qos: none broadcast: unsupported ronly: unsupported\n',
    'Full Zoning Database :\n',
    '    DB size: 220 bytes\n',
    '    Zonesets:2  Zones:2 Aliases: 0\n',
    'Active Zoning Database :\n',
    '    DB size: 80 bytes\n',
    '    Name: test-zs-test  Zonesets:1  Zones:1\n',
    'Status:\n']

zoning_status_basic = {'mode': 'basic', 'session': 'none'}

zoning_status_data_enhanced_nosess = [
    'VSAN: 303 default-zone: deny distribute: active only Interop: default\n',
    '    mode: enhanced merge-control: allow\n',
    '    session: none\n',
    '    hard-zoning: enabled broadcast: unsupported\n',
    '    smart-zoning: disabled\n',
    '    rscn-format: fabric-address\n',
    'Default zone:\n',
    '    qos: none broadcast: unsupported ronly: unsupported\n',
    'Full Zoning Database :\n',
    '    DB size: 220 bytes\n',
    '    Zonesets:2  Zones:2 Aliases: 0\n',
    'Active Zoning Database :\n',
    '    DB size: 80 bytes\n',
    '    Name: test-zs-test  Zonesets:1  Zones:1\n',
    'Status:\n']

zoning_status_enhanced_nosess = {'mode': 'enhanced', 'session': 'none'}

zoning_status_data_enhanced_sess = [
    'VSAN: 303 default-zone: deny distribute: active only Interop: default\n',
    '    mode: enhanced merge-control: allow\n',
    '    session: otherthannone\n',
    '    hard-zoning: enabled broadcast: unsupported\n',
    '    smart-zoning: disabled\n',
    '    rscn-format: fabric-address\n',
    'Default zone:\n',
    '    qos: none broadcast: unsupported ronly: unsupported\n',
    'Full Zoning Database :\n',
    '    DB size: 220 bytes\n',
    '    Zonesets:2  Zones:2 Aliases: 0\n',
    'Active Zoning Database :\n',
    '    DB size: 80 bytes\n',
    '    Name: test-zs-test  Zonesets:1  Zones:1\n',
    'Status:\n']

zoning_status_enhanced_sess = {'mode': 'enhanced', 'session': 'otherthannone'}

active_zoneset_multiple_zones = {
    'zones': {
        'openstack50060b0000c26604201900051ee8e329':
        ['50:06:0b:00:00:c2:66:04', '20:19:00:05:1e:e8:e3:29'],
        'openstack10000012345678902001009876543210':
        ['50:06:0b:00:00:c2:66:02', '20:19:00:05:1e:e8:e3:27']},
    'active_zone_config': 'OpenStack_Cfg'}

new_zone = {'openstack10000012345678902001009876543210':
            ['10:00:00:12:34:56:78:90', '20:01:00:98:76:54:32:10']}

new_zones = {'openstack10000012345678902001009876543210':
             ['10:00:00:12:34:56:78:90', '20:01:00:98:76:54:32:10'],
             'openstack10000011111111112001001111111111':
             ['10:00:00:11:11:11:11:11', '20:01:00:11:11:11:11:11']}

zone_names_to_delete = 'openstack50060b0000c26604201900051ee8e329'


class TestCiscoFCZoneClientCLI(cli.CiscoFCZoneClientCLI, test.TestCase):

    def setUp(self):
        super(TestCiscoFCZoneClientCLI, self).setUp()
        self.fabric_vsan = '303'

    # override some of the functions
    def __init__(self, *args, **kwargs):
        test.TestCase.__init__(self, *args, **kwargs)

    @mock.patch.object(cli.CiscoFCZoneClientCLI, '_get_switch_info')
    def test_get_active_zone_set(self, get_switch_info_mock):
        cmd_list = [ZoneConstant.GET_ACTIVE_ZONE_CFG, self.fabric_vsan,
                    ' | no-more']
        get_switch_info_mock.return_value = cfgactv
        active_zoneset_returned = self.get_active_zone_set()
        get_switch_info_mock.assert_called_once_with(cmd_list)
        self.assertDictEqual(active_zoneset, active_zoneset_returned)

    @mock.patch.object(cli.CiscoFCZoneClientCLI, '_run_ssh')
    def test_get_active_zone_set_ssh_error(self, run_ssh_mock):
        run_ssh_mock.side_effect = processutils.ProcessExecutionError
        self.assertRaises(c_exception.CiscoZoningCliException,
                          self.get_active_zone_set)

    @mock.patch.object(cli.CiscoFCZoneClientCLI, '_get_switch_info')
    def test_get_zoning_status_basic(self, get_zoning_status_mock):
        cmd_list = [ZoneConstant.GET_ZONE_STATUS, self.fabric_vsan]
        get_zoning_status_mock.return_value = zoning_status_data_basic
        zoning_status_returned = self.get_zoning_status()
        get_zoning_status_mock.assert_called_once_with(cmd_list)
        self.assertDictEqual(zoning_status_basic, zoning_status_returned)

    @mock.patch.object(cli.CiscoFCZoneClientCLI, '_get_switch_info')
    def test_get_zoning_status_enhanced_nosess(self, get_zoning_status_mock):
        cmd_list = [ZoneConstant.GET_ZONE_STATUS, self.fabric_vsan]
        get_zoning_status_mock.return_value =\
            zoning_status_data_enhanced_nosess
        zoning_status_returned = self.get_zoning_status()
        get_zoning_status_mock.assert_called_once_with(cmd_list)
        self.assertDictEqual(zoning_status_enhanced_nosess,
                             zoning_status_returned)

    @mock.patch.object(cli.CiscoFCZoneClientCLI, '_get_switch_info')
    def test_get_zoning_status_enhanced_sess(self, get_zoning_status_mock):
        cmd_list = [ZoneConstant.GET_ZONE_STATUS, self.fabric_vsan]
        get_zoning_status_mock.return_value = zoning_status_data_enhanced_sess
        zoning_status_returned = self.get_zoning_status()
        get_zoning_status_mock.assert_called_once_with(cmd_list)
        self.assertDictEqual(zoning_status_enhanced_sess,
                             zoning_status_returned)

    @mock.patch.object(cli.CiscoFCZoneClientCLI, '_get_switch_info')
    def test_get_nameserver_info(self, get_switch_info_mock):
        ns_info_list = []
        ns_info_list_expected = ['20:1a:00:05:1e:e8:e3:29']
        get_switch_info_mock.return_value = (switch_data)
        ns_info_list = self.get_nameserver_info()
        self.assertEqual(ns_info_list_expected, ns_info_list)

    @mock.patch.object(cli.CiscoFCZoneClientCLI, '_run_ssh')
    def test_get_nameserver_info_ssh_error(self, run_ssh_mock):
        run_ssh_mock.side_effect = processutils.ProcessExecutionError
        self.assertRaises(c_exception.CiscoZoningCliException,
                          self.get_nameserver_info)

    @mock.patch.object(cli.CiscoFCZoneClientCLI, '_run_ssh')
    def test__cfg_save(self, run_ssh_mock):
        cmd_list = ['copy', 'running-config', 'startup-config']
        self._cfg_save()
        run_ssh_mock.assert_called_once_with(cmd_list, True)

    @mock.patch.object(cli.CiscoFCZoneClientCLI, '_run_ssh')
    @mock.patch.object(time, 'sleep')
    def test__cfg_save_with_retry(self, mock_sleep, run_ssh_mock):
        cmd_list = ['copy', 'running-config', 'startup-config']
        run_ssh_mock.side_effect = [
            processutils.ProcessExecutionError,
            ('', None)
        ]

        self._cfg_save()

        self.assertEqual(2, run_ssh_mock.call_count)
        run_ssh_mock.assert_has_calls([
            mock.call(cmd_list, True),
            mock.call(cmd_list, True)
        ])

    @mock.patch.object(cli.CiscoFCZoneClientCLI, '_run_ssh')
    @mock.patch.object(time, 'sleep')
    def test__cfg_save_with_error(self, mock_sleep, run_ssh_mock):
        cmd_list = ['copy', 'running-config', 'startup-config']
        run_ssh_mock.side_effect = processutils.ProcessExecutionError

        self.assertRaises(processutils.ProcessExecutionError, self._cfg_save)

        expected_num_calls = 5
        expected_calls = []
        for i in range(expected_num_calls):
            expected_calls.append(mock.call(cmd_list, True))

        self.assertEqual(expected_num_calls, run_ssh_mock.call_count)
        run_ssh_mock.assert_has_calls(expected_calls)

    @mock.patch.object(cli.CiscoFCZoneClientCLI, '_run_ssh')
    def test__get_switch_info(self, run_ssh_mock):
        cmd_list = [ZoneConstant.FCNS_SHOW, self.fabric_vsan]
        nsshow_list = [nsshow]
        run_ssh_mock.return_value = (Stream(nsshow), Stream())
        switch_data = self._get_switch_info(cmd_list)
        self.assertEqual(nsshow_list, switch_data)
        run_ssh_mock.assert_called_once_with(cmd_list, True)

    @mock.patch.object(cli.CiscoFCZoneClientCLI, '_ssh_execute')
    @mock.patch.object(cli.CiscoFCZoneClientCLI, '_cfg_save')
    def test__update_zones_add(self, cfg_save_mock, ssh_execute_mock):
        self.update_zones(new_zone, False, self.fabric_vsan,
                          ZoneConstant.ZONE_ADD, active_zoneset_multiple_zones,
                          zoning_status_basic)
        ssh_cmd = [['conf'],
                   ['zoneset', 'name', 'OpenStack_Cfg', 'vsan',
                    self.fabric_vsan],
                   ['zone', 'name',
                    'openstack10000012345678902001009876543210'],
                   ['member', 'pwwn', '10:00:00:12:34:56:78:90'],
                   ['member', 'pwwn', '20:01:00:98:76:54:32:10'],
                   ['end']]

        self.assertEqual(1, cfg_save_mock.call_count)
        ssh_execute_mock.assert_called_once_with(ssh_cmd, True, 1)

    @mock.patch.object(cli.CiscoFCZoneClientCLI, '_ssh_execute')
    @mock.patch.object(cli.CiscoFCZoneClientCLI, '_cfg_save')
    def test__update_zones_remove(self, cfg_save_mock, ssh_execute_mock):
        self.update_zones(new_zone, False, self.fabric_vsan,
                          ZoneConstant.ZONE_REMOVE,
                          active_zoneset_multiple_zones,
                          zoning_status_basic)
        ssh_cmd = [['conf'],
                   ['zoneset', 'name', 'OpenStack_Cfg', 'vsan',
                    self.fabric_vsan],
                   ['zone', 'name',
                    'openstack10000012345678902001009876543210'],
                   ['no', 'member', 'pwwn', '10:00:00:12:34:56:78:90'],
                   ['no', 'member', 'pwwn', '20:01:00:98:76:54:32:10'],
                   ['end']]

        self.assertEqual(1, cfg_save_mock.call_count)
        ssh_execute_mock.assert_called_once_with(ssh_cmd, True, 1)

    def test__parse_ns_output(self):
        return_wwn_list = []
        expected_wwn_list = ['20:1a:00:05:1e:e8:e3:29']
        return_wwn_list = self._parse_ns_output(switch_data)
        self.assertEqual(expected_wwn_list, return_wwn_list)


class TestCiscoFCZoneClientCLISSH(test.TestCase):

    def setUp(self):
        super(TestCiscoFCZoneClientCLISSH, self).setUp()
        self.client = cli.CiscoFCZoneClientCLI(None, None, None, None, None)
        self.client.sshpool = mock.MagicMock()
        self.mock_ssh = self.client.sshpool.item().__enter__()

    @mock.patch('oslo_concurrency.processutils.ssh_execute')
    def test__run_ssh(self, mock_execute):
        mock_execute.return_value = 'ssh output'
        ret = self.client._run_ssh(['cat', 'foo'])
        self.assertEqual('ssh output', ret)
        mock_execute.assert_called_once_with(self.mock_ssh,
                                             'cat foo',
                                             check_exit_code=True)

    @mock.patch('oslo_concurrency.processutils.ssh_execute')
    def test__run_ssh_with_error(self, mock_execute):
        mock_execute.side_effect = processutils.ProcessExecutionError()
        self.assertRaises(processutils.ProcessExecutionError,
                          self.client._run_ssh,
                          ['cat', 'foo'])


class Channel(object):
    def recv_exit_status(self):
        return 0


class Stream(object):
    def __init__(self, buffer=''):
        self.buffer = buffer
        self.channel = Channel()

    def readlines(self):
        return self.buffer

    def splitlines(self):
        return self.buffer.splitlines()

    def close(self):
        pass

    def flush(self):
        self.buffer = ''



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from alembic.runtime import migration as alembic_migration
from migrate import exceptions as migrate_exceptions
from migrate.versioning import api as migrate_api
from oslotest import base as test_base

from cinder.db import migration
from cinder.db.sqlalchemy import api as db_api


class TestDBSync(test_base.BaseTestCase):

    def test_db_sync_legacy_version(self):
        """We don't allow users to request legacy versions."""
        self.assertRaises(ValueError, migration.db_sync, '402')

    @mock.patch.object(migration, '_upgrade_alembic')
    @mock.patch.object(migration, '_init_alembic_on_legacy_database')
    @mock.patch.object(migration, '_is_database_under_alembic_control')
    @mock.patch.object(migration, '_is_database_under_migrate_control')
    @mock.patch.object(migration, '_find_alembic_conf')
    @mock.patch.object(migration, '_find_migrate_repo')
    @mock.patch.object(db_api, 'get_engine')
    def _test_db_sync(
        self, has_migrate, has_alembic, mock_get_engine, mock_find_repo,
        mock_find_conf, mock_is_migrate, mock_is_alembic, mock_init,
        mock_upgrade,
    ):
        mock_is_migrate.return_value = has_migrate
        mock_is_alembic.return_value = has_alembic
        migration.db_sync()
        mock_get_engine.assert_called_once_with()
        mock_find_repo.assert_called_once_with()
        mock_find_conf.assert_called_once_with()
        mock_find_conf.return_value.set_main_option.assert_called_once_with(
            'sqlalchemy.url', str(mock_get_engine.return_value.url),
        )
        mock_is_migrate.assert_called_once_with(
            mock_get_engine.return_value, mock_find_repo.return_value)
        if has_migrate:
            mock_is_alembic.assert_called_once_with(
                mock_get_engine.return_value)
        else:
            mock_is_alembic.assert_not_called()

        # we should only attempt the upgrade of the remaining
        # sqlalchemy-migrate-based migrations and fake apply of the initial
        # alembic migrations if sqlalchemy-migrate is in place but alembic
        # hasn't been used yet
        if has_migrate and not has_alembic:
            mock_init.assert_called_once_with(
                mock_get_engine.return_value,
                mock_find_repo.return_value, mock_find_conf.return_value)
        else:
            mock_init.assert_not_called()

        # however, we should always attempt to upgrade the requested migration
        # to alembic
        mock_upgrade.assert_called_once_with(
            mock_get_engine.return_value, mock_find_conf.return_value, None)

    def test_db_sync_new_deployment(self):
        """Mimic a new deployment without existing sqlalchemy-migrate cruft."""
        has_migrate = False
        has_alembic = False
        self._test_db_sync(has_migrate, has_alembic)

    def test_db_sync_with_existing_migrate_database(self):
        """Mimic a deployment currently managed by sqlalchemy-migrate."""
        has_migrate = True
        has_alembic = False
        self._test_db_sync(has_migrate, has_alembic)

    def test_db_sync_with_existing_alembic_database(self):
        """Mimic a deployment that's already switched to alembic."""
        has_migrate = True
        has_alembic = True
        self._test_db_sync(has_migrate, has_alembic)


@mock.patch.object(alembic_migration.MigrationContext, 'configure')
@mock.patch.object(migrate_api, 'db_version')
@mock.patch.object(migration, '_is_database_under_alembic_control')
@mock.patch.object(migration, '_is_database_under_migrate_control')
@mock.patch.object(db_api, 'get_engine')
@mock.patch.object(migration, '_find_migrate_repo')
class TestDBVersion(test_base.BaseTestCase):

    def test_db_version_migrate(
        self, mock_find_repo, mock_get_engine, mock_is_migrate,
        mock_is_alembic, mock_migrate_version, mock_m_context_configure,
    ):
        """Database is controlled by sqlalchemy-migrate."""
        mock_is_migrate.return_value = True
        mock_is_alembic.return_value = False
        ret = migration.db_version()
        self.assertEqual(mock_migrate_version.return_value, ret)
        mock_find_repo.assert_called_once_with()
        mock_get_engine.assert_called_once_with()
        mock_is_migrate.assert_called_once()
        mock_is_alembic.assert_called_once()
        mock_migrate_version.assert_called_once_with(
            mock_get_engine.return_value, mock_find_repo.return_value)
        mock_m_context_configure.assert_not_called()

    def test_db_version_alembic(
        self, mock_find_repo, mock_get_engine, mock_is_migrate,
        mock_is_alembic, mock_migrate_version, mock_m_context_configure,
    ):
        """Database is controlled by alembic."""
        mock_is_migrate.return_value = False
        mock_is_alembic.return_value = True
        ret = migration.db_version()
        mock_m_context = mock_m_context_configure.return_value
        self.assertEqual(
            mock_m_context.get_current_revision.return_value,
            ret,
        )
        mock_find_repo.assert_called_once_with()
        mock_get_engine.assert_called_once_with()
        mock_is_migrate.assert_called_once()
        mock_is_alembic.assert_called_once()
        mock_migrate_version.assert_not_called()
        mock_m_context_configure.assert_called_once()

    def test_db_version_not_controlled(
        self, mock_find_repo, mock_get_engine, mock_is_migrate,
        mock_is_alembic, mock_migrate_version, mock_m_context_configure,
    ):
        """Database is not controlled."""
        mock_is_migrate.return_value = False
        mock_is_alembic.return_value = False
        ret = migration.db_version()
        self.assertIsNone(ret)
        mock_find_repo.assert_called_once_with()
        mock_get_engine.assert_called_once_with()
        mock_is_migrate.assert_called_once()
        mock_is_alembic.assert_called_once()
        mock_migrate_version.assert_not_called()
        mock_m_context_configure.assert_not_called()


class TestDatabaseUnderVersionControl(test_base.BaseTestCase):

    @mock.patch.object(migrate_api, 'db_version')
    def test__is_database_under_migrate_control__true(self, mock_db_version):
        ret = migration._is_database_under_migrate_control('engine', 'repo')
        self.assertTrue(ret)
        mock_db_version.assert_called_once_with('engine', 'repo')

    @mock.patch.object(migrate_api, 'db_version')
    def test__is_database_under_migrate_control__false(self, mock_db_version):
        mock_db_version.side_effect = \
            migrate_exceptions.DatabaseNotControlledError()
        ret = migration._is_database_under_migrate_control('engine', 'repo')
        self.assertFalse(ret)
        mock_db_version.assert_called_once_with('engine', 'repo')

    @mock.patch.object(alembic_migration.MigrationContext, 'configure')
    def test__is_database_under_alembic_control__true(self, mock_configure):
        context = mock_configure.return_value
        context.get_current_revision.return_value = 'foo'
        engine = mock.MagicMock()
        ret = migration._is_database_under_alembic_control(engine)
        self.assertTrue(ret)
        context.get_current_revision.assert_called_once_with()

    @mock.patch.object(alembic_migration.MigrationContext, 'configure')
    def test__is_database_under_alembic_control__false(self, mock_configure):
        context = mock_configure.return_value
        context.get_current_revision.return_value = None
        engine = mock.MagicMock()
        ret = migration._is_database_under_alembic_control(engine)
        self.assertFalse(ret)
        context.get_current_revision.assert_called_once_with()
